import logging
import typing as t

import numpy as np
import pandas as pd

from sarus_data_spec.constants import MAX_MAX_MULT
from sarus_data_spec.dataspec_validator.signature import SarusBoundSignature
from sarus_data_spec.protobuf.utilities import unwrap
import sarus_data_spec.typing as st

from ..external_op import ExternalOpImplementation

logger = logging.getLogger(__name__)

try:
    from sarus_statistics.ops.bounds.op import BoundOp
    from sarus_statistics.ops.corr.op import CorrOp
    from sarus_statistics.ops.histograms.op import CountOp, HistogramOp
    from sarus_statistics.ops.max_multiplicity.op import MaxMultiplicityOp
    from sarus_statistics.ops.mean.op import MeanOp
    from sarus_statistics.ops.median.op import MedianOp
    from sarus_statistics.ops.std.op import StdOp
    from sarus_statistics.ops.sum.op import SumOp
    from sarus_statistics.protobuf.multiplicity_pb2 import (
        MultiplicityParameters,
    )
except ModuleNotFoundError as e:
    if "sarus_statistics" not in str(e):
        raise

try:
    from sarus_differential_privacy.query import ComposedPrivateQuery
except ModuleNotFoundError as e:
    if "sarus_differential_privacy" not in str(e):
        raise

try:
    from sarus_query_builder.builders.bounds_builder import (
        simple_bounds_builder,
    )
    from sarus_query_builder.builders.composed_builder import (
        simple_composed_builder,
    )
    from sarus_query_builder.builders.corr_builder import (
        corr_builder as simple_corr_builder,
    )
    from sarus_query_builder.builders.max_multiplicity_builder import (
        simple_max_multiplicity_builder,
    )
    from sarus_query_builder.builders.mean_builder import mean_builder
    from sarus_query_builder.builders.median_builder import median_builder
    from sarus_query_builder.builders.standard_mechanisms_builder import (
        laplace_builder,
    )
    from sarus_query_builder.builders.std_builder import std_builder
    from sarus_query_builder.builders.sum_builder import sum_builder
    from sarus_query_builder.core.core import OptimizableQueryBuilder
    from sarus_query_builder.protobuf.query_pb2 import GenericTask, Query
except ModuleNotFoundError as e:
    if "sarus_query_builder" not in str(e):
        raise


DEFAULT_MAX_MAX_MULT = 1
NUMERIC_TYPES = ('integer', 'float', 'boolean')


class pd_shape_dp(ExternalOpImplementation):
    _transform_id = "pandas.PD_SHAPE_DP"
    _non_dp_equivalent_id = "pandas.PD_SHAPE"

    def is_dp(self, signature: SarusBoundSignature) -> bool:
        return True

    async def private_queries_and_task(
        self, signature: SarusBoundSignature
    ) -> t.Tuple[t.List[st.PrivateQuery], st.Task]:
        """Return the PrivateQueries summarizing DP characteristics."""
        parent_ds = signature["this"].static_value()
        epsilon, delta = await get_budget(signature)

        parent_schema = await parent_ds.manager().async_schema(parent_ds)
        max_max_mult = int(
            parent_schema.properties().get(MAX_MAX_MULT, DEFAULT_MAX_MAX_MULT)
        )

        builder = simple_composed_builder(
            parent_ds,
            [
                simple_max_multiplicity_builder(
                    parent_ds,
                    Query(
                        max_multiplicity=Query.MaxMultiplicity(
                            max_max_multiplicity=max_max_mult
                        )
                    ),
                ),
                laplace_builder(
                    parent_ds,
                    Query(laplace_mechanism=Query.LaplaceMechanism()),
                ),
            ],
        )
        tasks = builder.build_query(
            builder.target([(epsilon, delta)], (0, epsilon))
        )
        query = builder.private_query(tasks)
        composed_query = t.cast(ComposedPrivateQuery, query)
        return list(composed_query.all_subqueries()), tasks

    async def call_dp(self, signature: SarusBoundSignature) -> t.Any:
        """Implementation of DP shape.

        A DP implementation receives additional arguments compared to a
        standard external implementation:
            - `budget`: a list of sp.Scalar.PrivacyParams.Point
                object containing each an epsilon and a delta values
            - `seed`: an integer used to parametrize random number generators
            - `pe`: the protected entity used by `sarus_statistics` primitives
        """
        # Evaluate arguments
        parent_ds = signature.static_kwargs()["this"]
        (dataframe, budget, seed) = await signature.collect_args()

        # Get QB task parametrization
        _, tasks = await self.private_queries_and_task(signature)

        epsilon = budget[0].epsilon
        n_rows, n_cols = dataframe.shape

        # Compute DP value
        tasks = [unwrap(subtask) for subtask in tasks.subtasks]
        max_mult_task, shape_task = t.cast(
            MultiplicityParameters, tasks[0]
        ), t.cast(GenericTask, tasks[1])
        random_generator = np.random.default_rng(abs(seed))
        max_mul = MaxMultiplicityOp(
            parent_ds,
            epsilon,  # parameter for quantiles
            max_mult_task.noise_user_count,
            max_mult_task.noise_multiplicity,
            max_mult_task.max_max_multiplicity,
        ).value(random_generator)

        n_rows = CountOp(
            parent_ds,
            noise=shape_task.parameters['noise'],
        ).value(None, max_mul, random_generator)

        dp_shape = (n_rows, n_cols)
        return dp_shape


class pd_sum_dp(ExternalOpImplementation):
    _transform_id = "pandas.PD_SUM_DP"
    _non_dp_equivalent_id = "pandas.PD_SUM"

    def is_dp(self, signature: SarusBoundSignature) -> bool:
        """`axis = 0` and `numeric_only` is `True`"""
        if signature["this"].python_type() == str(pd.Series):
            return True

        axis = signature["axis"].static_value()
        numeric_only = signature["numeric_only"].static_value()
        return (axis == 0) and (numeric_only is True)

    async def private_queries_and_task(
        self, signature: SarusBoundSignature
    ) -> t.Tuple[t.List[st.PrivateQuery], st.Task]:
        parent_ds = signature["this"].static_value()
        epsilon, delta = await get_budget(signature)

        parent_schema = await parent_ds.manager().async_schema(parent_ds)
        max_max_mult = int(
            parent_schema.properties().get(MAX_MAX_MULT, DEFAULT_MAX_MAX_MULT)
        )
        n_cols = len(parent_schema.data_type().children())

        max_mult_builder: OptimizableQueryBuilder = (
            simple_max_multiplicity_builder(
                parent_ds,
                Query(
                    max_multiplicity=Query.MaxMultiplicity(
                        max_max_multiplicity=max_max_mult
                    )
                ),
            )
        )

        column_bounds_sum_builder: OptimizableQueryBuilder = (
            simple_composed_builder(
                parent_ds,
                [
                    simple_bounds_builder(
                        parent_ds, Query(bounds=Query.Bounds())
                    ),
                    sum_builder(parent_ds, Query(sum=Query.Sum())),
                ],
            )
        )

        builder = simple_composed_builder(
            parent_ds,
            [max_mult_builder] + n_cols * [column_bounds_sum_builder],
        )
        tasks = builder.build_query(
            builder.target([(epsilon, delta)], (0, epsilon))
        )
        query = builder.private_query(tasks)
        composed_query = t.cast(ComposedPrivateQuery, query)
        return list(composed_query.all_subqueries()), tasks

    async def call_dp(self, signature: SarusBoundSignature) -> t.Any:
        # Evaluate arguments
        parent_ds: st.Dataset = signature["this"].static_value()
        dataframe: pd.DataFrame = await parent_ds.async_to(pd.DataFrame)
        budget_value = await signature["budget"].collect()
        seed_value = await signature["seed"].collect()
        skipna = await signature["skipna"].collect()

        _, tasks = await self.private_queries_and_task(signature)

        epsilon = budget_value[0].epsilon
        # Compute DP value
        tasks = [unwrap(subtask) for subtask in tasks.subtasks]
        max_mult_task, sum_tasks = tasks[0], tasks[1:]
        random_generator = np.random.default_rng(abs(seed_value))
        max_mul = MaxMultiplicityOp(
            parent_ds,
            epsilon,  # parameter for quantiles
            max_mult_task.noise_user_count,
            max_mult_task.noise_multiplicity,
            max_mult_task.max_max_multiplicity,
        ).value(random_generator)

        sum_dp_dict = {}
        for index, column_name in enumerate(dataframe.columns):
            subtasks = [unwrap(task) for task in sum_tasks[index].subtasks]
            bounds_parameters, sum_parameters = t.cast(
                GenericTask, subtasks[0]
            ), t.cast(GenericTask, subtasks[1])
            column_type = (
                parent_ds.schema()
                .data_type()
                .children()[column_name]
                .protobuf()
            )
            if column_type.WhichOneof('type') == 'optional':
                if not skipna:
                    sum_dp_dict[column_name] = np.nan
                    continue
                column_type = column_type.optional.type
            if column_type.WhichOneof('type') not in NUMERIC_TYPES:
                continue

            # TODO parent_ds doesn't have a size
            # if parent_ds.size().children()[column_name].size() < min_count:
            #     sum_dp_dict[column_name] = np.nan
            #     continue

            bounds = BoundOp(
                parent_ds,
                bounds_parameters.parameters['noise'],
            ).value(column_name, max_mul, random_generator)
            sum_op = SumOp(
                parent_ds,
                sum_parameters.parameters['noise'],
            )
            sum_dp_dict[column_name] = sum_op.value(
                column_name, max_mul, bounds, random_generator
            )

        if signature["this"].python_type() == str(pd.DataFrame):
            sum_dp = pd.Series(sum_dp_dict)
        else:
            sum_dp = list(sum_dp_dict.values()).pop()

        return sum_dp


class pd_mean_dp(ExternalOpImplementation):
    _transform_id = "pandas.PD_MEAN_DP"
    _non_dp_equivalent_id = "pandas.PD_MEAN"

    def is_dp(self, signature: SarusBoundSignature) -> bool:
        """`axis = 0` and `numeric_only` is `True`"""
        if signature["this"].python_type() == str(pd.Series):
            return True

        axis = signature["axis"].static_value()
        numeric_only = signature["numeric_only"].static_value()
        return (axis == 0) and (numeric_only is True)

    async def private_queries_and_task(
        self, signature: SarusBoundSignature
    ) -> t.Tuple[t.List[st.PrivateQuery], st.Task]:
        parent_ds = signature["this"].static_value()
        epsilon, delta = await get_budget(signature)

        parent_schema = await parent_ds.manager().async_schema(parent_ds)
        max_max_mult = int(
            parent_schema.properties().get(MAX_MAX_MULT, DEFAULT_MAX_MAX_MULT)
        )
        n_cols = len(parent_schema.data_type().children())

        max_mult_builder: OptimizableQueryBuilder = (
            simple_max_multiplicity_builder(
                parent_ds,
                Query(
                    max_multiplicity=Query.MaxMultiplicity(
                        max_max_multiplicity=max_max_mult
                    )
                ),
            )
        )
        column_bounds_mean_builder: OptimizableQueryBuilder = (
            simple_composed_builder(
                parent_ds,
                [
                    simple_bounds_builder(
                        parent_ds, Query(bounds=Query.Bounds())
                    ),
                    mean_builder(parent_ds, Query(mean=Query.Mean())),
                ],
            )
        )
        builder = simple_composed_builder(
            parent_ds,
            [max_mult_builder] + n_cols * [column_bounds_mean_builder],
        )
        tasks = builder.build_query(
            builder.target([(epsilon, delta)], (0, epsilon))
        )
        query = builder.private_query(tasks)
        composed_query = t.cast(ComposedPrivateQuery, query)
        return list(composed_query.all_subqueries()), tasks

    async def call_dp(self, signature: SarusBoundSignature) -> t.Any:
        # Evaluate arguments
        parent_ds: st.Dataset = signature["this"].static_value()
        dataframe: pd.DataFrame = await parent_ds.async_to(pd.DataFrame)
        budget_value = await signature["budget"].collect()
        seed_value = await signature["seed"].collect()
        skipna = await signature["skipna"].collect()
        pe = (await signature.admin_data()).to_pandas()

        _, tasks = await self.private_queries_and_task(signature)

        tasks = [unwrap(subtask) for subtask in tasks.subtasks]
        max_mult_task, mean_tasks = tasks[0], tasks[1:]

        # Compute DP value
        random_generator = np.random.default_rng(abs(seed_value))
        epsilon = budget_value[0].epsilon
        max_mul = MaxMultiplicityOp(
            parent_ds,
            epsilon,  # parameter for quantiles
            max_mult_task.noise_user_count,
            max_mult_task.noise_multiplicity,
            max_mult_task.max_max_multiplicity,
        ).value(random_generator)

        mean_dp_dict = {}
        for index, column_name in enumerate(dataframe.columns):
            df = dataframe[[column_name]].join(pe)
            subtasks = [unwrap(task) for task in mean_tasks[index].subtasks]
            bounds_parameters, mean_parameters = t.cast(
                GenericTask, subtasks[0]
            ), t.cast(GenericTask, subtasks[1])
            column_type = (
                parent_ds.schema()
                .data_type()
                .children()[column_name]
                .protobuf()
            )
            if column_type.WhichOneof('type') == 'optional':
                if not skipna:
                    mean_dp_dict[column_name] = np.nan
                    continue
                column_type = column_type.optional.type
                df = df.dropna()
            if column_type.WhichOneof('type') not in NUMERIC_TYPES:
                continue

            bounds = BoundOp(
                parent_ds,
                bounds_parameters.parameters['noise'],
            ).value(column_name, max_mul, random_generator)
            mean_op = MeanOp(
                parent_ds,
                mean_parameters.parameters['noise'],
            )
            mean_dp_dict[column_name] = mean_op.value(
                column_name, max_mul, bounds, random_generator
            )

        if signature["this"].python_type() == str(pd.DataFrame):
            mean_dp = pd.Series(mean_dp_dict)
        else:
            mean_dp = list(mean_dp_dict.values()).pop()

        return mean_dp


class pd_median_dp(ExternalOpImplementation):
    _transform_id = "pandas.PD_MEDIAN_DP"
    _non_dp_equivalent_id = "pandas.PD_MEDIAN"

    def is_dp(self, signature: SarusBoundSignature) -> bool:
        """`axis = 0` and `numeric_only` is `True`"""
        if signature["this"].python_type() == str(pd.Series):
            return True

        axis = signature["axis"].static_value()
        numeric_only = signature["numeric_only"].static_value()
        return axis == 0 and numeric_only is True

    async def private_queries_and_task(
        self, signature: SarusBoundSignature
    ) -> t.Tuple[t.List[st.PrivateQuery], st.Task]:
        parent_ds = signature["this"].static_value()
        epsilon, delta = await get_budget(signature)

        parent_schema = await parent_ds.manager().async_schema(parent_ds)
        max_max_mult = int(
            parent_schema.properties().get(MAX_MAX_MULT, DEFAULT_MAX_MAX_MULT)
        )
        n_cols = len(parent_schema.data_type().children())

        max_mult_builder: OptimizableQueryBuilder = (
            simple_max_multiplicity_builder(
                parent_ds,
                Query(
                    max_multiplicity=Query.MaxMultiplicity(
                        max_max_multiplicity=max_max_mult
                    )
                ),
            )
        )
        column_bounds_median_builder: OptimizableQueryBuilder = (
            simple_composed_builder(
                parent_ds,
                [
                    simple_bounds_builder(
                        parent_ds, Query(bounds=Query.Bounds())
                    ),
                    median_builder(parent_ds, Query(median=Query.Median())),
                ],
            )
        )

        builder = simple_composed_builder(
            parent_ds,
            [max_mult_builder] + n_cols * [column_bounds_median_builder],
        )
        tasks = builder.build_query(
            builder.target([(epsilon, delta)], (0, epsilon))
        )
        query = builder.private_query(tasks)
        composed_query = t.cast(ComposedPrivateQuery, query)
        return list(composed_query.all_subqueries()), tasks

    async def call_dp(self, signature: SarusBoundSignature) -> t.Any:
        # Evaluate arguments
        parent_ds: st.Dataset = signature["this"].static_value()
        dataframe: pd.DataFrame = await parent_ds.async_to(pd.DataFrame)
        budget_value = await signature["budget"].collect()
        seed_value = await signature["seed"].collect()
        skipna = await signature["skipna"].collect()
        pe = (await signature.admin_data()).to_pandas()

        _, tasks = await self.private_queries_and_task(signature)
        tasks = [unwrap(subtask) for subtask in tasks.subtasks]
        max_mult_task, median_tasks = tasks[0], tasks[1:]

        random_generator = np.random.default_rng(abs(seed_value))
        epsilon = budget_value[0].epsilon

        # Compute DP value
        max_mul = MaxMultiplicityOp(
            parent_ds,
            epsilon,  # parameter for quantiles
            max_mult_task.noise_user_count,
            max_mult_task.noise_multiplicity,
            max_mult_task.max_max_multiplicity,
        ).value(random_generator)

        median_dp_dict = {}
        for index, column_name in enumerate(dataframe.columns):
            df = dataframe[[column_name]].join(pe)
            subtasks = [unwrap(task) for task in median_tasks[index].subtasks]
            bounds_parameters, median_parameters = t.cast(
                GenericTask, subtasks[0]
            ), t.cast(GenericTask, subtasks[1])
            column_type = (
                parent_ds.schema()
                .data_type()
                .children()[column_name]
                .protobuf()
            )
            if column_type.WhichOneof('type') == 'optional':
                if not skipna:
                    median_dp_dict[column_name] = np.nan
                    continue
                column_type = column_type.optional.type
                df = df.dropna()
            if column_type.WhichOneof('type') not in NUMERIC_TYPES:
                continue

            bounds = BoundOp(
                parent_ds,
                bounds_parameters.parameters['noise'],
            ).value(column_name, max_mul, random_generator)
            median_op = MedianOp(
                parent_ds,
                median_parameters.parameters['noise'],
            )
            median_dp_dict[column_name] = median_op.value(
                column_name, max_mul, bounds, random_generator
            )

        if signature["this"].python_type() == str(pd.DataFrame):
            median_dp = pd.Series(median_dp_dict)
        else:
            median_dp = list(median_dp_dict.values()).pop()

        return median_dp


class pd_std_dp(ExternalOpImplementation):
    _transform_id = "pandas.PD_STD_DP"
    _non_dp_equivalent_id = "pandas.PD_STD"

    def is_dp(self, signature: SarusBoundSignature) -> bool:
        """`axis = 0` and `numeric_only` is `True`"""
        if signature["this"].python_type() == str(pd.Series):
            return True

        axis = signature["axis"].static_value()
        numeric_only = signature["numeric_only"].static_value()
        return axis == 0 and numeric_only is True

    async def private_queries_and_task(
        self, signature: SarusBoundSignature
    ) -> t.Tuple[t.List[st.PrivateQuery], st.Task]:
        parent_ds = signature["this"].static_value()
        epsilon, delta = await get_budget(signature)

        parent_schema = await parent_ds.manager().async_schema(parent_ds)
        max_max_mult = int(
            parent_schema.properties().get(MAX_MAX_MULT, DEFAULT_MAX_MAX_MULT)
        )
        n_cols = len(parent_schema.data_type().children())

        max_mult_builder: OptimizableQueryBuilder = (
            simple_max_multiplicity_builder(
                parent_ds,
                Query(
                    max_multiplicity=Query.MaxMultiplicity(
                        max_max_multiplicity=max_max_mult
                    )
                ),
            )
        )
        column_bounds_std_builder: OptimizableQueryBuilder = (
            simple_composed_builder(
                parent_ds,
                [
                    simple_bounds_builder(
                        parent_ds, Query(bounds=Query.Bounds())
                    ),
                    std_builder(parent_ds, Query(std=Query.Std())),
                ],
            )
        )

        builder = simple_composed_builder(
            parent_ds,
            [max_mult_builder] + n_cols * [column_bounds_std_builder],
        )
        tasks = builder.build_query(
            builder.target([(epsilon, delta)], (0, epsilon))
        )
        query = builder.private_query(tasks)
        composed_query = t.cast(ComposedPrivateQuery, query)
        return list(composed_query.all_subqueries()), tasks

    async def call_dp(self, signature: SarusBoundSignature) -> t.Any:
        # Evaluate arguments
        parent_ds: st.Dataset = signature["this"].static_value()
        dataframe: pd.DataFrame = await parent_ds.async_to(pd.DataFrame)
        budget_value = await signature["budget"].collect()
        seed_value = await signature["seed"].collect()
        skipna = await signature["skipna"].collect()
        pe = (await signature.admin_data()).to_pandas()

        _, tasks = await self.private_queries_and_task(signature)
        tasks = [unwrap(subtask) for subtask in tasks.subtasks]
        max_mult_task, std_tasks = tasks[0], tasks[1:]

        epsilon = budget_value[0].epsilon
        random_generator = np.random.default_rng(abs(seed_value))

        # Compute DP value
        max_mul = MaxMultiplicityOp(
            parent_ds,
            epsilon,  # parameter for quantiles
            max_mult_task.noise_user_count,
            max_mult_task.noise_multiplicity,
            max_mult_task.max_max_multiplicity,
        ).value(random_generator)

        std_dp_dict = {}
        for index, column_name in enumerate(dataframe.columns):
            df = dataframe[[column_name]].join(pe)
            subtasks = [unwrap(task) for task in std_tasks[index].subtasks]
            bounds_parameters, std_parameters = t.cast(
                GenericTask, subtasks[0]
            ), t.cast(GenericTask, subtasks[1])
            column_type = (
                parent_ds.schema()
                .data_type()
                .children()[column_name]
                .protobuf()
            )
            if column_type.WhichOneof('type') == 'optional':
                if not skipna:
                    std_dp_dict[column_name] = np.nan
                    continue
                column_type = column_type.optional.type
                df = df.dropna()
            if column_type.WhichOneof('type') not in NUMERIC_TYPES:
                continue

            bounds = BoundOp(
                parent_ds,
                bounds_parameters.parameters['noise'],
            ).value(column_name, max_mul, random_generator)
            std_op = StdOp(
                parent_ds,
                std_parameters.parameters['noise_mean'],
                std_parameters.parameters['noise_square'],
                std_parameters.parameters['noise_count'],
            )
            std_dp_dict[column_name] = std_op.value(
                column_name, max_mul, bounds, random_generator
            )

        if signature["this"].python_type() == str(pd.DataFrame):
            std_dp = pd.Series(std_dp_dict)
        else:
            std_dp = list(std_dp_dict.values()).pop()

        return std_dp


class pd_count_dp(ExternalOpImplementation):
    _transform_id = "pandas.PD_COUNT_DP"
    _non_dp_equivalent_id = "pandas.PD_COUNT"

    def is_dp(self, signature: SarusBoundSignature) -> bool:
        """`axis = 0` and `numeric_only` is `True`"""
        axis = signature["axis"].static_value()
        return bool(axis == 0)

    async def private_queries_and_task(
        self, signature: SarusBoundSignature
    ) -> t.Tuple[t.List[st.PrivateQuery], st.Task]:
        parent_ds = signature["this"].static_value()
        epsilon, delta = await get_budget(signature)

        parent_schema = await parent_ds.manager().async_schema(parent_ds)
        max_max_mult = int(
            parent_schema.properties().get(MAX_MAX_MULT, DEFAULT_MAX_MAX_MULT)
        )
        n_cols = len(parent_schema.data_type().children())

        max_mult_builder: OptimizableQueryBuilder = (
            simple_max_multiplicity_builder(
                parent_ds,
                Query(
                    max_multiplicity=Query.MaxMultiplicity(
                        max_max_multiplicity=max_max_mult
                    )
                ),
            )
        )
        column_counts_builder = laplace_builder(
            parent_ds, Query(laplace_mechanism=Query.LaplaceMechanism())
        )
        builders_list: t.List[OptimizableQueryBuilder] = [max_mult_builder]
        builders_list.extend(n_cols * [column_counts_builder])
        builder = simple_composed_builder(parent_ds, builders_list)
        tasks = builder.build_query(
            builder.target([(epsilon, delta)], (0, epsilon))
        )
        query = builder.private_query(tasks)
        composed_query = t.cast(ComposedPrivateQuery, query)
        return list(composed_query.all_subqueries()), tasks

    async def call_dp(self, signature: SarusBoundSignature) -> t.Any:
        # Evaluate arguments
        parent_ds: st.Dataset = signature["this"].static_value()
        dataframe: pd.DataFrame = await parent_ds.async_to(pd.DataFrame)
        budget_value = await signature["budget"].collect()
        seed_value = await signature["seed"].collect()

        _, tasks = await self.private_queries_and_task(signature)
        tasks = [unwrap(subtask) for subtask in tasks.subtasks]
        max_mult_task, shape_task = tasks[0], tasks[1:]

        epsilon = budget_value[0].epsilon
        random_generator = np.random.default_rng(abs(seed_value))

        # Compute DP value
        max_mul = MaxMultiplicityOp(
            parent_ds,
            epsilon,  # parameter for quantiles
            max_mult_task.noise_user_count,
            max_mult_task.noise_multiplicity,
            max_mult_task.max_max_multiplicity,
        ).value(random_generator)

        count_dp_dict = {}
        for index, column_name in enumerate(dataframe.columns):
            count_parameters = shape_task[index]
            count_dp_dict[column_name] = CountOp(
                parent_ds,
                count_parameters.parameters['noise'],
            ).value(column_name, max_mul, random_generator)

        if signature["this"].python_type() == str(pd.DataFrame):
            count_dp = pd.Series(count_dp_dict)
        else:
            count_dp = list(count_dp_dict.values()).pop()
        return count_dp


class pd_value_counts_dp(ExternalOpImplementation):
    _transform_id: str = "pandas.PD_VALUE_COUNTS_DP"
    _non_dp_equivalent_id = "pandas.PD_VALUE_COUNTS"

    def is_dp(self, signature: SarusBoundSignature) -> bool:
        return signature["this"].python_type() == str(pd.Series)

    async def private_queries_and_task(
        self, signature: SarusBoundSignature
    ) -> t.Tuple[t.List[st.PrivateQuery], st.Task]:
        """Return the PrivateQueries summarizing DP characteristics."""
        parent_ds = signature["this"].static_value()
        epsilon, delta = await get_budget(signature)

        parent_schema = await parent_ds.manager().async_schema(parent_ds)
        max_max_mult = int(
            parent_schema.properties().get(MAX_MAX_MULT, DEFAULT_MAX_MAX_MULT)
        )
        n_cols = len(parent_schema.data_type().children())

        max_mult_builder: OptimizableQueryBuilder = (
            simple_max_multiplicity_builder(
                parent_ds,
                Query(
                    max_multiplicity=Query.MaxMultiplicity(
                        max_max_multiplicity=max_max_mult
                    )
                ),
            )
        )
        column_counts_builder = laplace_builder(
            parent_ds, Query(laplace_mechanism=Query.LaplaceMechanism())
        )
        builders_list: t.List[OptimizableQueryBuilder] = [max_mult_builder]
        builders_list.extend(n_cols * [column_counts_builder])
        builder = simple_composed_builder(parent_ds, builders_list)
        tasks = builder.build_query(
            builder.target([(epsilon, delta)], (0, epsilon))
        )
        query = builder.private_query(tasks)
        composed_query = t.cast(ComposedPrivateQuery, query)
        return list(composed_query.all_subqueries()), tasks

    async def call_dp(self, signature: SarusBoundSignature) -> t.Any:
        """Implementation of DP shape.

        A DP implementation receives additional arguments compared to a
        standard external implementation:
            - `budget`: a list of sp.Scalar.PrivacyParams.Point
                object containing each an epsilon and a delta values
            - `seed`: an integer used to parametrize random number generators
            - `pe`: the protected entity used by `sarus_statistics` primitives
        """
        parent_ds: st.Dataset = signature["this"].static_value()
        dataframe: pd.DataFrame = await parent_ds.async_to(pd.DataFrame)
        budget_value = await signature["budget"].collect()
        seed_value = await signature["seed"].collect()
        sort = await signature["sort"].collect()
        ascending = await signature["ascending"].collect()
        normalize = await signature["normalize"].collect()
        dropna = await signature["dropna"].collect()

        _, tasks = await self.private_queries_and_task(signature)
        tasks = [unwrap(subtask) for subtask in tasks.subtasks]
        max_mult_task, shape_task = tasks[0], tasks[1:]

        epsilon = budget_value[0].epsilon
        random_generator = np.random.default_rng(abs(seed_value))

        # Compute DP value
        max_mul = MaxMultiplicityOp(
            parent_ds,
            epsilon,  # parameter for quantiles
            max_mult_task.noise_user_count,
            max_mult_task.noise_multiplicity,
            max_mult_task.max_max_multiplicity,
        ).value(random_generator)

        count_dp_dict = {}
        for index, column_name in enumerate(dataframe.columns):
            count_parameters = shape_task[index]

            histogram = t.cast(
                t.Dict[str, float],
                HistogramOp(
                    parent_ds,
                    noise=count_parameters.parameters['noise'],
                    sort=sort,
                    ascending=ascending,
                    normalize=normalize,
                    dropna=dropna,
                ).value(column_name, max_mul, random_generator),
            )
            count_dp_dict[column_name] = {
                key: round(count) for key, count in histogram.items()
            }

        # for now only on series
        count_dp = pd.Series(list(count_dp_dict.values())[0])
        return count_dp


class pd_corr_dp(ExternalOpImplementation):
    _transform_id = "pandas.PD_CORR_DP"
    _non_dp_equivalent_id = "pandas.PD_CORR"

    def is_dp(self, signature: SarusBoundSignature) -> bool:
        """
        `numeric_only` is `True`
        DataFrame
        `method = Pearson`
        `min_periods = 1`
        """
        method = signature["method"].static_value()
        min_periods = signature["min_periods"].static_value()
        is_dataframe = bool(
            signature["this"].python_type() == str(pd.DataFrame)
        )
        return is_dataframe and (min_periods == 1) and (method == 'pearson')

    async def private_queries_and_task(
        self, signature: SarusBoundSignature
    ) -> t.Tuple[t.List[st.PrivateQuery], st.Task]:
        parent_ds = signature["this"].static_value()
        epsilon, delta = await get_budget(signature)

        parent_schema = await parent_ds.manager().async_schema(parent_ds)
        max_max_mult = int(
            parent_schema.properties().get(MAX_MAX_MULT, DEFAULT_MAX_MAX_MULT)
        )

        max_mult_builder: OptimizableQueryBuilder = (
            simple_max_multiplicity_builder(
                parent_ds,
                Query(
                    max_multiplicity=Query.MaxMultiplicity(
                        max_max_multiplicity=max_max_mult
                    )
                ),
            )
        )

        corr_builder: OptimizableQueryBuilder = simple_corr_builder(
            parent_ds, Query(covariance=Query.Covariance())
        )

        builder = simple_composed_builder(
            parent_ds,
            [max_mult_builder, corr_builder],
        )
        tasks = builder.build_query(
            builder.target([(epsilon, delta)], (0, epsilon))
        )
        query = t.cast(ComposedPrivateQuery, builder.private_query(tasks))
        return list(query.all_subqueries()), tasks

    async def call_dp(self, signature: SarusBoundSignature) -> t.Any:
        # Evaluate arguments
        parent_ds: st.Dataset = signature["this"].static_value()
        budget_value = await signature["budget"].collect()
        seed_value = await signature["seed"].collect()

        _, tasks = await self.private_queries_and_task(signature)

        epsilon = budget_value[0].epsilon
        # Compute DP value
        tasks = [unwrap(subtask) for subtask in tasks.subtasks]
        max_mult_task, corr_task = tasks[0], tasks[1]
        random_generator = np.random.default_rng(abs(seed_value))

        max_mul = MaxMultiplicityOp(
            parent_ds,
            epsilon,  # parameter for quantiles
            max_mult_task.noise_user_count,
            max_mult_task.noise_multiplicity,
            max_mult_task.max_max_multiplicity,
        ).value(random_generator)

        corr_dp = CorrOp(
            parent_ds,
            corr_task.parameters['epsilon'],
            corr_task.parameters['dims'],
        ).value(max_mul, random_generator=random_generator)

        return corr_dp


async def get_budget(signature: SarusBoundSignature) -> t.Tuple[float, float]:
    """Retrieve (epsilon, delta) from signature"""
    budget = await signature["budget"].collect()
    if len(budget) != 1:
        raise NotImplementedError(
            "The PrivacyParams contains more than 1 point in the privacy "
            "profile."
        )

    epsilon = budget[0].epsilon
    delta = budget[0].delta
    if epsilon == 0.0:
        raise ValueError("`epsilon` should be greater than 0.")

    return epsilon, delta



# flake8: noqa
from .cluster import *
from .compose import *
from .decomposition import *
from .ensemble import *
from .impute import *
from .inspection import *
from .lib import *
from .linear_model import *
from .metrics import *
from .model_selection import *
from .pipeline import *
from .preprocessing import *
from .svm import *



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: sarus_data_spec/protobuf/status.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n%sarus_data_spec/protobuf/status.proto\x12\x0fsarus_data_spec\"\xe5\x05\n\x06Status\x12\x0c\n\x04uuid\x18\x01 \x01(\t\x12\x10\n\x08\x64\x61taspec\x18\x02 \x01(\t\x12\x0f\n\x07manager\x18\x03 \x01(\t\x12\x10\n\x08\x64\x61tetime\x18\x04 \x01(\t\x12<\n\x0btask_stages\x18\x05 \x03(\x0b\x32\'.sarus_data_spec.Status.TaskStagesEntry\x12;\n\nproperties\x18\x06 \x03(\x0b\x32\'.sarus_data_spec.Status.PropertiesEntry\x1aP\n\x0fTaskStagesEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12,\n\x05value\x18\x02 \x01(\x0b\x32\x1d.sarus_data_spec.Status.Stage:\x02\x38\x01\x1a\x31\n\x0fPropertiesEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01\x1a\x97\x03\n\x05Stage\x12\x38\n\x07pending\x18\x01 \x01(\x0b\x32%.sarus_data_spec.Status.Stage.PendingH\x00\x12>\n\nprocessing\x18\x02 \x01(\x0b\x32(.sarus_data_spec.Status.Stage.ProcessingH\x00\x12\x34\n\x05ready\x18\x03 \x01(\x0b\x32#.sarus_data_spec.Status.Stage.ReadyH\x00\x12\x34\n\x05\x65rror\x18\x04 \x01(\x0b\x32#.sarus_data_spec.Status.Stage.ErrorH\x00\x12\x41\n\nproperties\x18\x05 \x03(\x0b\x32-.sarus_data_spec.Status.Stage.PropertiesEntry\x1a\x31\n\x0fPropertiesEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01\x1a\t\n\x07Pending\x1a\x0c\n\nProcessing\x1a\x07\n\x05Ready\x1a\x07\n\x05\x45rrorB\x07\n\x05stageb\x06proto3')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'sarus_data_spec.protobuf.status_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  _STATUS_TASKSTAGESENTRY._options = None
  _STATUS_TASKSTAGESENTRY._serialized_options = b'8\001'
  _STATUS_PROPERTIESENTRY._options = None
  _STATUS_PROPERTIESENTRY._serialized_options = b'8\001'
  _STATUS_STAGE_PROPERTIESENTRY._options = None
  _STATUS_STAGE_PROPERTIESENTRY._serialized_options = b'8\001'
  _STATUS._serialized_start=59
  _STATUS._serialized_end=800
  _STATUS_TASKSTAGESENTRY._serialized_start=259
  _STATUS_TASKSTAGESENTRY._serialized_end=339
  _STATUS_PROPERTIESENTRY._serialized_start=341
  _STATUS_PROPERTIESENTRY._serialized_end=390
  _STATUS_STAGE._serialized_start=393
  _STATUS_STAGE._serialized_end=800
  _STATUS_STAGE_PROPERTIESENTRY._serialized_start=341
  _STATUS_STAGE_PROPERTIESENTRY._serialized_end=390
  _STATUS_STAGE_PENDING._serialized_start=750
  _STATUS_STAGE_PENDING._serialized_end=759
  _STATUS_STAGE_PROCESSING._serialized_start=761
  _STATUS_STAGE_PROCESSING._serialized_end=773
  _STATUS_STAGE_READY._serialized_start=775
  _STATUS_STAGE_READY._serialized_end=782
  _STATUS_STAGE_ERROR._serialized_start=784
  _STATUS_STAGE_ERROR._serialized_end=791
# @@protoc_insertion_point(module_scope)



from __future__ import annotations

import typing as t

import numpy as np
import pyarrow as pa

from sarus_data_spec.manager.async_utils import decoupled_async_iter
import sarus_data_spec.dataspec_validator.typing as sdvt
import sarus_data_spec.typing as st


class DataspecStaticChecker:
    def __init__(self, dataspec: st.DataSpec):
        self.dataspec = dataspec

    def is_dp_applicable(self, public_context: t.Collection[str]) -> bool:
        """Statically check if a DP transform is applicable in this position.

        This verification is common to all dataspecs and is true if:
            - the dataspec is transformed and its transform has an equivalent
            DP transform
            - the DP transform's required PEP arguments are PEP and aligned
            (i.e. same PEP token)
            - other dataspecs arguments are public
        """
        return False

    def is_dp(self) -> bool:
        """Checks if the transform is DP and compatible with the arguments."""
        return False

    def dp_transform(self) -> t.Optional[st.Transform]:
        """Return the dataspec's DP equivalent transform if existing."""
        return None

    def dp_equivalent(self) -> t.Optional[st.Transform]:
        """Return the dataspec's DP equivalent transform if existing."""
        raise NotImplementedError

    async def private_queries(self) -> t.List[st.PrivateQuery]:
        """Return the PrivateQueries summarizing DP characteristics."""
        if self.is_dp():
            raise NotImplementedError
        else:
            return []


class DatasetStaticChecker(DataspecStaticChecker):
    def __init__(self, dataset: st.Dataset):
        super().__init__(dataset)
        self.dataset = dataset

    async def schema(self) -> st.Schema:
        """Computes the schema of the dataspec"""
        raise NotImplementedError

    def pep_token(self, public_context: t.Collection[str]) -> t.Optional[str]:
        """Return a token if the output is PEP."""
        raise NotImplementedError

    def pep_kind(self) -> sdvt.PEPKind:
        raise NotImplementedError


class DatasetImplementation:
    def __init__(self, dataset: st.Dataset):
        self.dataset = dataset

    async def to_arrow(
        self, batch_size: int
    ) -> t.AsyncIterator[pa.RecordBatch]:
        raise NotImplementedError

    async def size(self) -> st.Size:
        raise NotImplementedError

    async def multiplicity(self) -> st.Multiplicity:
        raise NotImplementedError

    async def bounds(self) -> st.Bounds:
        raise NotImplementedError

    async def marginals(self) -> st.Marginals:
        raise NotImplementedError

    async def sql(
        self,
        query: t.Union[str, t.Dict[str, t.Any]],
        dialect: t.Optional[st.SQLDialect] = None,
        batch_size: int = 10000,
        result_type: t.Optional[st.Type] = None,
    ) -> t.AsyncIterator[pa.RecordBatch]:
        """It composes the query and it sends it to the parent."""
        raise NotImplementedError

    @staticmethod
    async def decoupled_async_iter(
        source: t.AsyncIterator[pa.RecordBatch], buffer_size: int = 100
    ) -> t.AsyncIterator[pa.RecordBatch]:
        return decoupled_async_iter(source=source, buffer_size=buffer_size)


class ScalarImplementation:
    def __init__(self, scalar: st.Scalar):
        self.scalar = scalar

    async def value(self) -> t.Any:
        raise NotImplementedError

    @staticmethod
    async def decoupled_async_iter(
        source: t.AsyncIterator[pa.RecordBatch], buffer_size: int = 100
    ) -> t.AsyncIterator[pa.RecordBatch]:
        return decoupled_async_iter(source=source, buffer_size=buffer_size)


async def _ensure_batch_correct(
    async_iterator: t.Union[
        t.AsyncIterator[pa.RecordBatch], t.AsyncIterator[pa.Array]
    ],
    func_to_apply: t.Callable,
    batch_size: int,
) -> t.AsyncIterator[pa.RecordBatch]:
    """Method that executes func_to_apply on each batch
    of the async_iterator but rather than directly returning
    the result, it accumulates them and returns them progressively
    so that each new batch has batch_size."""

    global_array = None
    async for batch in async_iterator:
        new_array = await func_to_apply(batch)
        if len(new_array) == batch_size and global_array is None:
            yield pa.RecordBatch.from_struct_array(
                new_array.take(
                    np.linspace(0, batch_size - 1, batch_size, dtype=int)
                )
            )
        elif global_array is not None:
            global_array = pa.concat_arrays([global_array, new_array])
            if len(global_array) < batch_size:
                continue
            else:
                # here cannot use array.slice because there
                # is a bug in the columns being copied
                # when we switch to record batch
                yield pa.RecordBatch.from_struct_array(
                    global_array.take(
                        np.linspace(0, batch_size - 1, batch_size, dtype=int)
                    )
                )
                global_array = global_array.take(
                    np.linspace(
                        batch_size,
                        len(global_array) - 1,
                        len(global_array) - batch_size,
                        dtype=int,
                    )
                )

        else:
            # initialize global_array
            global_array = new_array
            continue
    # handle remaining array: split it in

    if global_array is not None and len(global_array) > 0:
        while len(global_array) > 0:
            min_val = min(batch_size, len(global_array))
            indices = np.linspace(
                0, len(global_array) - 1, len(global_array), dtype=int
            )
            yield pa.RecordBatch.from_struct_array(
                global_array.take(indices[:min_val])
            )
            global_array = global_array.take(indices[min_val:])



from typing import Any, Literal, Optional, Union

import numpy as np

from sarus_data_spec.dataspec_validator.signature import (
    SarusParameter,
    SarusSignature,
    SarusSignatureValue,
)

from ..external_op import ExternalOpImplementation

try:
    from sklearn import decomposition
except ModuleNotFoundError:
    pass  # error message in typing.py


class sk_pca(ExternalOpImplementation):
    _transform_id = "sklearn.SK_PCA"
    _signature = SarusSignature(
        SarusParameter(
            name="n_components",
            annotation=Optional[Union[int, float, Literal["mle"]]],
            default=None,
        ),
        SarusParameter(
            name="copy",
            annotation=bool,
            default=True,
        ),
        SarusParameter(
            name="whiten",
            annotation=bool,
            default=False,
        ),
        SarusParameter(
            name="svd_solver",
            annotation=Literal['auto', 'full', 'arpack', 'randomized'],
            default='auto',
        ),
        SarusParameter(
            name="tol",
            annotation=float,
            default=0.0,
        ),
        SarusParameter(
            name="iterated_power",
            annotation=Union[int, Literal["auto"]],
            default='auto',
        ),
        SarusParameter(
            name="n_oversamples",
            annotation=int,
            default=10,
        ),
        SarusParameter(
            name="power_iteration_normalizer",
            annotation=Literal['auto', 'QR', 'LU', 'none'],
            default='auto',
        ),
        SarusParameter(
            name="random_state",
            annotation=Optional[Union[int, np.random.RandomState]],
            default=None,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        kwargs = signature.collect_kwargs()
        return decomposition.PCA(**kwargs)



from typing import Any, Callable, Dict, Iterable, List, Optional, Tuple, Union

import numpy as np

from sarus_data_spec.dataspec_validator.signature import (
    SarusParameter,
    SarusSignature,
    SarusSignatureValue,
)

from .external_op import ExternalOpImplementation

try:
    from sklearn.base import BaseEstimator
    from sklearn.model_selection import BaseCrossValidator
    import skopt
except ModuleNotFoundError:
    BaseEstimator = Any
    BaseCrossValidator = Any


class skopt_bayes_search_cv(ExternalOpImplementation):
    _transform_id = "skopt.SKOPT_BAYES_SEARCH_CV"
    _signature = SarusSignature(
        SarusParameter(
            name="estimator",
            annotation=BaseEstimator,
        ),
        SarusParameter(
            name="search_spaces",
            annotation=Union[Dict, List[Union[Dict, Tuple]]],
        ),
        SarusParameter(
            name="n_iter",
            annotation=int,
            default=50,
        ),
        SarusParameter(
            name="optimizer_kwargs",
            annotation=Optional[Dict],
            default=None,
        ),
        SarusParameter(
            name="scoring",
            annotation=Optional[Union[str, Callable]],
            default=None,
        ),
        SarusParameter(
            name="fit_params",
            annotation=Optional[Dict],
            default=None,
        ),
        SarusParameter(
            name="n_jobs",
            annotation=int,
            default=1,
            predicate=lambda x: bool(x < 4),
        ),
        SarusParameter(
            name="n_points",
            annotation=int,
            default=1,
        ),
        SarusParameter(
            name="pre_dispatch",
            annotation=Optional[Union[int, str]],
            default="2*n_jobs",
        ),
        SarusParameter(
            name="iid",
            annotation=bool,
            default=True,
        ),
        SarusParameter(
            name="cv",
            annotation=Optional[Union[int, BaseCrossValidator, Iterable]],
            default=None,
        ),
        SarusParameter(
            name="refit",
            annotation=bool,
            default=True,
        ),
        SarusParameter(
            name="verbose",
            annotation=int,
            default=0,
        ),
        SarusParameter(
            name="random_state",
            annotation=Union[int, np.random.RandomState],
            default=None,
        ),
        SarusParameter(
            name="error_score",
            annotation=Union[str, float],
            default="raise",
        ),
        SarusParameter(
            name="return_train_score",
            annotation=bool,
            default=False,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        kwargs = signature.collect_kwargs()
        return skopt.BayesSearchCV(**kwargs)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: sarus_data_spec/protobuf/statistics.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n)sarus_data_spec/protobuf/statistics.proto\x12\x0fsarus_data_spec\"\xbb\x18\n\nStatistics\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x30\n\x04null\x18\x03 \x01(\x0b\x32 .sarus_data_spec.Statistics.NullH\x00\x12\x30\n\x04unit\x18\x04 \x01(\x0b\x32 .sarus_data_spec.Statistics.UnitH\x00\x12\x36\n\x07\x62oolean\x18\x05 \x01(\x0b\x32#.sarus_data_spec.Statistics.BooleanH\x00\x12\x36\n\x07integer\x18\x06 \x01(\x0b\x32#.sarus_data_spec.Statistics.IntegerH\x00\x12\x30\n\x04\x65num\x18\x07 \x01(\x0b\x32 .sarus_data_spec.Statistics.EnumH\x00\x12\x32\n\x05\x66loat\x18\x08 \x01(\x0b\x32!.sarus_data_spec.Statistics.FloatH\x00\x12\x30\n\x04text\x18\t \x01(\x0b\x32 .sarus_data_spec.Statistics.TextH\x00\x12\x32\n\x05\x62ytes\x18\n \x01(\x0b\x32!.sarus_data_spec.Statistics.BytesH\x00\x12\x34\n\x06struct\x18\x0b \x01(\x0b\x32\".sarus_data_spec.Statistics.StructH\x00\x12\x32\n\x05union\x18\x0c \x01(\x0b\x32!.sarus_data_spec.Statistics.UnionH\x00\x12\x38\n\x08optional\x18\r \x01(\x0b\x32$.sarus_data_spec.Statistics.OptionalH\x00\x12\x30\n\x04list\x18\x0e \x01(\x0b\x32 .sarus_data_spec.Statistics.ListH\x00\x12\x32\n\x05\x61rray\x18\x0f \x01(\x0b\x32!.sarus_data_spec.Statistics.ArrayH\x00\x12\x38\n\x08\x64\x61tetime\x18\x10 \x01(\x0b\x32$.sarus_data_spec.Statistics.DatetimeH\x00\x12,\n\x02id\x18\x11 \x01(\x0b\x32\x1e.sarus_data_spec.Statistics.IdH\x00\x12>\n\x0b\x63onstrained\x18\x12 \x01(\x0b\x32\'.sarus_data_spec.Statistics.ConstrainedH\x00\x12\x30\n\x04\x64\x61te\x18\x13 \x01(\x0b\x32 .sarus_data_spec.Statistics.DateH\x00\x12\x30\n\x04time\x18\x14 \x01(\x0b\x32 .sarus_data_spec.Statistics.TimeH\x00\x12\x38\n\x08\x64uration\x18\x15 \x01(\x0b\x32$.sarus_data_spec.Statistics.DurationH\x00\x12?\n\nproperties\x18\x02 \x03(\x0b\x32+.sarus_data_spec.Statistics.PropertiesEntry\x1a\x31\n\x0fPropertiesEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01\x1a*\n\x04Null\x12\x0c\n\x04size\x18\x01 \x01(\x03\x12\x14\n\x0cmultiplicity\x18\x02 \x01(\x01\x1a*\n\x04Unit\x12\x0c\n\x04size\x18\x01 \x01(\x03\x12\x14\n\x0cmultiplicity\x18\x02 \x01(\x01\x1a\x62\n\x07\x42oolean\x12\x33\n\x0c\x64istribution\x18\x01 \x01(\x0b\x32\x1d.sarus_data_spec.Distribution\x12\x0c\n\x04size\x18\x02 \x01(\x03\x12\x14\n\x0cmultiplicity\x18\x03 \x01(\x01\x1a(\n\x02Id\x12\x0c\n\x04size\x18\x01 \x01(\x03\x12\x14\n\x0cmultiplicity\x18\x02 \x01(\x01\x1a\x62\n\x07Integer\x12\x33\n\x0c\x64istribution\x18\x01 \x01(\x0b\x32\x1d.sarus_data_spec.Distribution\x12\x0c\n\x04size\x18\x02 \x01(\x03\x12\x14\n\x0cmultiplicity\x18\x03 \x01(\x01\x1a_\n\x04\x45num\x12\x33\n\x0c\x64istribution\x18\x01 \x01(\x0b\x32\x1d.sarus_data_spec.Distribution\x12\x0c\n\x04size\x18\x02 \x01(\x03\x12\x14\n\x0cmultiplicity\x18\x03 \x01(\x01\x1a`\n\x05\x46loat\x12\x33\n\x0c\x64istribution\x18\x01 \x01(\x0b\x32\x1d.sarus_data_spec.Distribution\x12\x0c\n\x04size\x18\x02 \x01(\x03\x12\x14\n\x0cmultiplicity\x18\x03 \x01(\x01\x1ap\n\x04Text\x12\x33\n\x0c\x64istribution\x18\x01 \x01(\x0b\x32\x1d.sarus_data_spec.Distribution\x12\x0f\n\x07\x65xample\x18\x02 \x01(\t\x12\x0c\n\x04size\x18\x03 \x01(\x03\x12\x14\n\x0cmultiplicity\x18\x04 \x01(\x01\x1a+\n\x05\x42ytes\x12\x0c\n\x04size\x18\x01 \x01(\x03\x12\x14\n\x0cmultiplicity\x18\x02 \x01(\x01\x1a\xbc\x01\n\x06Struct\x12\x38\n\x06\x66ields\x18\x01 \x03(\x0b\x32(.sarus_data_spec.Statistics.Struct.Field\x12\x0c\n\x04size\x18\x02 \x01(\x03\x12\x0c\n\x04name\x18\x03 \x01(\t\x12\x14\n\x0cmultiplicity\x18\x04 \x01(\x01\x1a\x46\n\x05\x46ield\x12\x0c\n\x04name\x18\x01 \x01(\t\x12/\n\nstatistics\x18\x02 \x01(\x0b\x32\x1b.sarus_data_spec.Statistics\x1a\xba\x01\n\x05Union\x12\x37\n\x06\x66ields\x18\x01 \x03(\x0b\x32\'.sarus_data_spec.Statistics.Union.Field\x12\x0c\n\x04size\x18\x02 \x01(\x03\x12\x0c\n\x04name\x18\x03 \x01(\t\x12\x14\n\x0cmultiplicity\x18\x04 \x01(\x01\x1a\x46\n\x05\x46ield\x12\x0c\n\x04name\x18\x01 \x01(\t\x12/\n\nstatistics\x18\x02 \x01(\x0b\x32\x1b.sarus_data_spec.Statistics\x1am\n\x08Optional\x12/\n\nstatistics\x18\x01 \x01(\x0b\x32\x1b.sarus_data_spec.Statistics\x12\x0c\n\x04size\x18\x02 \x01(\x03\x12\x0c\n\x04name\x18\x03 \x01(\t\x12\x14\n\x0cmultiplicity\x18\x04 \x01(\x01\x1a\x90\x01\n\x04List\x12/\n\nstatistics\x18\x01 \x01(\x0b\x32\x1b.sarus_data_spec.Statistics\x12\x33\n\x0c\x64istribution\x18\x02 \x01(\x0b\x32\x1d.sarus_data_spec.Distribution\x12\x0c\n\x04size\x18\x03 \x01(\x03\x12\x14\n\x0cmultiplicity\x18\x04 \x01(\x01\x1a\x92\x01\n\x05\x41rray\x12/\n\nstatistics\x18\x01 \x01(\x0b\x32\x1b.sarus_data_spec.Statistics\x12\x34\n\rdistributions\x18\x02 \x03(\x0b\x32\x1d.sarus_data_spec.Distribution\x12\x0c\n\x04size\x18\x03 \x01(\x03\x12\x14\n\x0cmultiplicity\x18\x04 \x01(\x01\x1a\x63\n\x08\x44\x61tetime\x12\x33\n\x0c\x64istribution\x18\x01 \x01(\x0b\x32\x1d.sarus_data_spec.Distribution\x12\x0c\n\x04size\x18\x02 \x01(\x03\x12\x14\n\x0cmultiplicity\x18\x03 \x01(\x01\x1a_\n\x04\x44\x61te\x12\x33\n\x0c\x64istribution\x18\x01 \x01(\x0b\x32\x1d.sarus_data_spec.Distribution\x12\x0c\n\x04size\x18\x02 \x01(\x03\x12\x14\n\x0cmultiplicity\x18\x03 \x01(\x01\x1a_\n\x04Time\x12\x33\n\x0c\x64istribution\x18\x01 \x01(\x0b\x32\x1d.sarus_data_spec.Distribution\x12\x0c\n\x04size\x18\x02 \x01(\x03\x12\x14\n\x0cmultiplicity\x18\x03 \x01(\x01\x1a\x63\n\x08\x44uration\x12\x33\n\x0c\x64istribution\x18\x01 \x01(\x0b\x32\x1d.sarus_data_spec.Distribution\x12\x0c\n\x04size\x18\x02 \x01(\x03\x12\x14\n\x0cmultiplicity\x18\x03 \x01(\x01\x1a\x62\n\x0b\x43onstrained\x12/\n\nstatistics\x18\x01 \x01(\x0b\x32\x1b.sarus_data_spec.Statistics\x12\x0c\n\x04size\x18\x02 \x01(\x03\x12\x14\n\x0cmultiplicity\x18\x03 \x01(\x01\x42\x0c\n\nstatistics\"\x93\x07\n\x0c\x44istribution\x12\x36\n\x06\x64ouble\x18\x01 \x01(\x0b\x32$.sarus_data_spec.Distribution.DoubleH\x00\x12\x38\n\x07integer\x18\x02 \x01(\x0b\x32%.sarus_data_spec.Distribution.IntegerH\x00\x12\x38\n\x07\x62oolean\x18\x03 \x01(\x0b\x32%.sarus_data_spec.Distribution.BooleanH\x00\x12\x32\n\x04\x65num\x18\x04 \x01(\x0b\x32\".sarus_data_spec.Distribution.EnumH\x00\x12\x41\n\nproperties\x18\x05 \x03(\x0b\x32-.sarus_data_spec.Distribution.PropertiesEntry\x1a\x31\n\x0fPropertiesEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01\x1a\x8b\x01\n\x06\x44ouble\x12:\n\x06points\x18\x01 \x03(\x0b\x32*.sarus_data_spec.Distribution.Double.Point\x12\x0b\n\x03min\x18\x02 \x01(\x01\x12\x0b\n\x03max\x18\x03 \x01(\x01\x1a+\n\x05Point\x12\r\n\x05value\x18\x02 \x01(\x01\x12\x13\n\x0bprobability\x18\x03 \x01(\x01\x1a\x8d\x01\n\x07Integer\x12;\n\x06points\x18\x01 \x03(\x0b\x32+.sarus_data_spec.Distribution.Integer.Point\x12\x0b\n\x03min\x18\x02 \x01(\x03\x12\x0b\n\x03max\x18\x03 \x01(\x03\x1a+\n\x05Point\x12\r\n\x05value\x18\x01 \x01(\x03\x12\x13\n\x0bprobability\x18\x02 \x01(\x01\x1a\x81\x01\n\x07\x42oolean\x12;\n\x06points\x18\x01 \x03(\x0b\x32+.sarus_data_spec.Distribution.Boolean.Point\x1a\x39\n\x05Point\x12\x0c\n\x04name\x18\x01 \x01(\x08\x12\r\n\x05value\x18\x02 \x01(\x03\x12\x13\n\x0bprobability\x18\x03 \x01(\x01\x1a{\n\x04\x45num\x12\x38\n\x06points\x18\x01 \x03(\x0b\x32(.sarus_data_spec.Distribution.Enum.Point\x1a\x39\n\x05Point\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\x01\x12\x13\n\x0bprobability\x18\x03 \x01(\x01\x42\x0e\n\x0c\x64istributionb\x06proto3')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'sarus_data_spec.protobuf.statistics_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  _STATISTICS_PROPERTIESENTRY._options = None
  _STATISTICS_PROPERTIESENTRY._serialized_options = b'8\001'
  _DISTRIBUTION_PROPERTIESENTRY._options = None
  _DISTRIBUTION_PROPERTIESENTRY._serialized_options = b'8\001'
  _STATISTICS._serialized_start=63
  _STATISTICS._serialized_end=3194
  _STATISTICS_PROPERTIESENTRY._serialized_start=1164
  _STATISTICS_PROPERTIESENTRY._serialized_end=1213
  _STATISTICS_NULL._serialized_start=1215
  _STATISTICS_NULL._serialized_end=1257
  _STATISTICS_UNIT._serialized_start=1259
  _STATISTICS_UNIT._serialized_end=1301
  _STATISTICS_BOOLEAN._serialized_start=1303
  _STATISTICS_BOOLEAN._serialized_end=1401
  _STATISTICS_ID._serialized_start=1403
  _STATISTICS_ID._serialized_end=1443
  _STATISTICS_INTEGER._serialized_start=1445
  _STATISTICS_INTEGER._serialized_end=1543
  _STATISTICS_ENUM._serialized_start=1545
  _STATISTICS_ENUM._serialized_end=1640
  _STATISTICS_FLOAT._serialized_start=1642
  _STATISTICS_FLOAT._serialized_end=1738
  _STATISTICS_TEXT._serialized_start=1740
  _STATISTICS_TEXT._serialized_end=1852
  _STATISTICS_BYTES._serialized_start=1854
  _STATISTICS_BYTES._serialized_end=1897
  _STATISTICS_STRUCT._serialized_start=1900
  _STATISTICS_STRUCT._serialized_end=2088
  _STATISTICS_STRUCT_FIELD._serialized_start=2018
  _STATISTICS_STRUCT_FIELD._serialized_end=2088
  _STATISTICS_UNION._serialized_start=2091
  _STATISTICS_UNION._serialized_end=2277
  _STATISTICS_UNION_FIELD._serialized_start=2018
  _STATISTICS_UNION_FIELD._serialized_end=2088
  _STATISTICS_OPTIONAL._serialized_start=2279
  _STATISTICS_OPTIONAL._serialized_end=2388
  _STATISTICS_LIST._serialized_start=2391
  _STATISTICS_LIST._serialized_end=2535
  _STATISTICS_ARRAY._serialized_start=2538
  _STATISTICS_ARRAY._serialized_end=2684
  _STATISTICS_DATETIME._serialized_start=2686
  _STATISTICS_DATETIME._serialized_end=2785
  _STATISTICS_DATE._serialized_start=2787
  _STATISTICS_DATE._serialized_end=2882
  _STATISTICS_TIME._serialized_start=2884
  _STATISTICS_TIME._serialized_end=2979
  _STATISTICS_DURATION._serialized_start=2981
  _STATISTICS_DURATION._serialized_end=3080
  _STATISTICS_CONSTRAINED._serialized_start=3082
  _STATISTICS_CONSTRAINED._serialized_end=3180
  _DISTRIBUTION._serialized_start=3197
  _DISTRIBUTION._serialized_end=4112
  _DISTRIBUTION_PROPERTIESENTRY._serialized_start=1164
  _DISTRIBUTION_PROPERTIESENTRY._serialized_end=1213
  _DISTRIBUTION_DOUBLE._serialized_start=3556
  _DISTRIBUTION_DOUBLE._serialized_end=3695
  _DISTRIBUTION_DOUBLE_POINT._serialized_start=3652
  _DISTRIBUTION_DOUBLE_POINT._serialized_end=3695
  _DISTRIBUTION_INTEGER._serialized_start=3698
  _DISTRIBUTION_INTEGER._serialized_end=3839
  _DISTRIBUTION_INTEGER_POINT._serialized_start=3796
  _DISTRIBUTION_INTEGER_POINT._serialized_end=3839
  _DISTRIBUTION_BOOLEAN._serialized_start=3842
  _DISTRIBUTION_BOOLEAN._serialized_end=3971
  _DISTRIBUTION_BOOLEAN_POINT._serialized_start=3914
  _DISTRIBUTION_BOOLEAN_POINT._serialized_end=3971
  _DISTRIBUTION_ENUM._serialized_start=3973
  _DISTRIBUTION_ENUM._serialized_end=4096
  _DISTRIBUTION_ENUM_POINT._serialized_start=4039
  _DISTRIBUTION_ENUM_POINT._serialized_end=4096
# @@protoc_insertion_point(module_scope)



from typing import Any, Literal, Optional, Union

import numpy as np
import pandas as pd

from sarus_data_spec.dataspec_validator.signature import (
    SarusParameter,
    SarusSignature,
    SarusSignatureValue,
)

from ..external_op import ExternalOpImplementation

try:
    from sklearn import impute
except ModuleNotFoundError:
    pass  # error message in typing.py


class sk_simple_imputer(ExternalOpImplementation):
    _transform_id = "sklearn.SK_SIMPLE_IMPUTER"
    _signature = SarusSignature(
        SarusParameter(
            name="missing_values",
            annotation=Optional[
                Union[int, float, str, type(np.nan), type(pd.NA)]
            ],
            default=np.nan,
        ),
        SarusParameter(
            name="strategy",
            annotation=Literal["mean", "median", "most_frequent", "constant"],
            default="mean",
        ),
        SarusParameter(
            name="fill_value",
            annotation=Optional[Union[str, np.number]],
            default=None,
        ),
        SarusParameter(
            name="verbose",
            annotation=int,
            default=0,
        ),
        SarusParameter(
            name="copy",
            annotation=bool,
            default=True,
        ),
        SarusParameter(
            name="add_indicator",
            annotation=bool,
            default=False,
        ),
        SarusParameter(
            name="keep_empty_features",
            annotation=bool,
            default=False,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        kwargs = signature.collect_kwargs()
        return impute.SimpleImputer(**kwargs)



import typing as t

from sarus_data_spec.manager.ops.base import ScalarImplementation


class RandomSeed(ScalarImplementation):
    async def value(self) -> t.Any:
        assert self.scalar.is_random_seed()
        return self.scalar.protobuf().spec.random_seed.value



from typing import Any, Optional

from sarus_data_spec.dataspec_validator.signature import (
    SarusParameter,
    SarusSignature,
    SarusSignatureValue,
)

from ..external_op import ExternalOpImplementation

try:
    from sklearn import linear_model
except ModuleNotFoundError:
    pass  # error message in typing.py


class sk_linear_regression(ExternalOpImplementation):
    _transform_id = "sklearn.SK_LINEAR_REGRESSION"
    _signature = SarusSignature(
        SarusParameter(
            name="fit_intercept",
            annotation=bool,
            default=True,
        ),
        SarusParameter(
            name="copy_X",
            annotation=bool,
            default=True,
        ),
        SarusParameter(
            name="n_jobs",
            annotation=Optional[int],
            default=None,
            predicate=lambda x: x is None,
        ),
        SarusParameter(
            name="positive",
            annotation=bool,
            default=False,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        kwargs = signature.collect_kwargs()
        return linear_model.LinearRegression(**kwargs)



from typing import Final

from sarus_data_spec.attribute import Attribute
from sarus_data_spec.context import push_global_context
from sarus_data_spec.dataset import Dataset
from sarus_data_spec.scalar import Scalar
from sarus_data_spec.status import Status
from sarus_data_spec.transform import Transform
from sarus_data_spec.variant_constraint import VariantConstraint

"""A library to manage Sarus datasets"""
# pylint: disable=unused-variable

PACKAGE_NAME: Final[str] = 'sarus_data_spec'
VERSION: Final[str] = '3.5.15'

try:
    import sarus_data_spec.context.worker as sw

    push_global_context(sw.WorkerContext())
except ModuleNotFoundError as exception:
    if exception.name == 'sarus_data_spec.context.worker':
        pass  # The worker context is absent from the public release
    else:
        raise exception



# pylint: disable=too-many-return-statements
from __future__ import annotations

from itertools import zip_longest
import typing as t
import warnings

from scipy.stats import binom

from sarus_data_spec.path import straight_path
import sarus_data_spec.typing as st

try:
    from sarus_sql.ast_transform.utils import split_quote
    import sqlalchemy as sa
except ModuleNotFoundError:
    warnings.warn('sqlalchemy not installed')

T = t.TypeVar("T")


def access_table_or_col(
    metadata_or_table: t.Any, key: str, create_cte: bool = True
) -> t.Any:
    """access child from an metadata_or_table.
    metadata_or_table: t.Union[
        sa.MetaData, sa.Table, t.Dict[str, sa.Table],  sa.sql.Selectable]

    if sa.sql.Selectable:
        returns the underlying sa.sql.expression.ColumnElement
    if sa.Table:
        returns sa.sql.expression.ColumnElement using the key.
    if create_cte is True access column using a column table expression (cte).
        returns a sa.sql.Selectable. In this cases the column is aliased
        by sqlalchemy.
    if metadata:
        returns sa.Table if no schema.
        if metadata was generated with schemas, it becomes of the form:
        metadata.tables = {'schema.table_name': table}
        which is a problem. It is thus split into:
        {'schema': {'table_name': table}}
        to make it usable.
    if t.Dict[str, sa.Table]:
        returns sa.Table using from key.
    """
    if isinstance(metadata_or_table, (sa.sql.Selectable, sa.Table)):
        if create_cte:
            return sa.select(
                # label('') tells SQLAlchemy to handle the alias for such col
                # without it will generate an alias with the same name as col
                # with it will assign a non-ambigous and allowed alias
                getattr(metadata_or_table.exported_columns, key).label(''),
            ).cte()
        if isinstance(metadata_or_table, sa.sql.Selectable):
            return metadata_or_table.exported_columns._index[0]
        return getattr(metadata_or_table.c, key)

    if isinstance(metadata_or_table, sa.MetaData):
        result: t.Dict[str, t.Any] = {}
        for child_key, child_value in metadata_or_table.tables.items():
            first_key, *rest = split_quote(child_key, '')
            if rest:
                if first_key not in result:
                    result[first_key] = {'.'.join(rest): child_value}
                else:
                    result[first_key]['.'.join(rest)] = child_value
            else:
                result[first_key] = child_value
        return result[key]
    else:
        # it is a dict on the form
        return metadata_or_table[key]


def grouped(
    iterable: t.Iterable[T], n: int = 2
) -> t.Iterable[t.Tuple[T, ...]]:
    """s -> (s0,s1,s2,...sn-1), (sn,sn+1,sn+2,...s2n-1), ..."""
    return zip_longest(*[iter(iterable)] * n)


class SqlOperations:
    """It stores sql queries to be executed for bigdata computations.
    List of sql operations to be performed for bigdata. Results are
    stored in a dict: {path: {operation: result}}
    """

    def __init__(self, group_by: int = 1) -> None:
        self.queries: t.List[
            t.Tuple[sa.sql.ClauseElement, t.Optional[st.Type]]
        ] = []
        self.group_by = group_by

    def append(
        self,
        query: t.Optional[sa.sql.ClauseElement],
        result_type: t.Optional[st.Type] = None,
    ) -> None:
        if query is not None:
            self.queries.append((query, result_type))

    def extend(self, queries: t.Optional[SqlOperations]) -> None:
        if queries:
            self.queries.extend(queries.queries)

    async def execute(
        self,
        dataset: st.Dataset,
    ) -> t.Dict[st.Path, t.Dict[str, t.Any]]:
        """Execute all queries and reformat it as:
        {path: {operation: result}}. If types are provided they are passed to
        async_sql.
        """
        if len(self.queries) == 0:
            return {}

        results = []
        for grouped_queries in grouped(self.queries, self.group_by):
            queries = [query for query in grouped_queries if query]
            result_types: t.List[t.Optional[st.Type]] = [
                query[1] for query in queries
            ]
            assert len(set(result_types)) == 1
            union_query = sa.union_all(
                *[statement for statement, result_type in queries]
            )

            rendered = sqlalchemy_query_to_string(union_query)
            query_result = await dataset.manager().async_sql(
                dataset, rendered, result_type=result_types[0]
            )
            results.extend(
                [
                    tuple(raw.values())
                    async for batch in query_result
                    for raw in batch.to_pylist()
                ]
            )

        # Reformat results as {path: {operation: result}}
        result_dict: t.Dict[st.Path, t.Dict[str, t.Any]] = {}
        str2path: t.Dict[str, st.Path] = {}
        for path_str, operation, *result in results:
            result_length = len(result)

            if path_str in str2path:
                path = str2path[path_str]
            else:
                splitted_path = (
                    [path_str]
                    if not path_str
                    else list(split_quote(path_str, ''))
                )
                path = straight_path(splitted_path)
                str2path[path_str] = path

            if path not in result_dict:
                if result_length == 1:
                    result_dict[path] = {operation: result[0]}
                elif result_length == 2:
                    result_dict[path] = {operation: {result[0]: result[1]}}
                else:
                    result_dict[path] = {operation: result}
            else:
                path_dict = result_dict[path]
                if result_length == 1:
                    if operation not in path_dict:
                        path_dict[operation] = result[0]
                    else:
                        if isinstance(path_dict[operation], t.List):
                            path_dict[operation].append(result[0])
                        else:
                            path_dict[operation] = [
                                path_dict[operation],
                                result[0],
                            ]

                elif result_length == 2:
                    if operation not in path_dict:
                        path_dict[operation] = {}
                    path_dict[operation][result[0]] = result[1]
                else:
                    path_dict[operation] = result
        return result_dict


def path_to_quoted_string(path: st.Path) -> str:
    """It transforms a path into a quoted string.
    eg:
    (my_schema, my.table) -> "my_schema"."my.table"

    where (my_schema, my.table) is the tuple representation of the st.Path

    Used to create sql queries and table objects manipulated by sqlachemy
    metadata.
    """
    return '.'.join(f'"{path}"' for path in path.to_strings_list()[0])


def sqlalchemy_query_to_string(sa_query: sa.sql.ClauseElement) -> str:
    """Compiles the sqlalchemy query into a string"""
    return str(sa_query.compile(compile_kwargs={"literal_binds": True}))


def find_optimal_multiplier_fraction(
    n_start: int,
    min_sample: int,
    p_optimal: float = 0.9999,
    tol: float = 1e-4,
) -> float:
    """
    For a random subsample, this function calculates the optimal multiplier
    for a fraction derived from the ratio of the minimum sample size
    to the original sample size. The multiplier is determined such that it
    reach the probability (p_optimal) of obtaining
    a random subsample with a size greater than the minimum sample size.

    Parameters
    ----------
    n_start : int
        Original sample size.
    min_sample : int
        Minimum sample size or target sample size.
    p_optimal : float, optional
        Minimum probability threshold that the subsample size is at least
        min_sample. Default is 0.9999.
    tol : float, optional
        Tolerance for the bisection search. Default is 1e-4.

    Returns
    -------
    float
        Optimal multiplier for the fraction (min_sample/n_start).

    Note
    ----
    If the fraction is greater than or equal to 1, the function returns 1.

    """
    if min_sample == 0 or n_start == 0:
        return 0.0

    fraction = min_sample / n_start
    if fraction >= 1:
        return 1.0

    lower: float = 0.0
    upper = 1 / fraction
    while upper - lower > tol:
        mid = (upper + lower) / 2
        proba = 1 - binom.cdf(min_sample, n_start, mid * fraction)
        if proba < p_optimal:
            lower = mid
        else:
            upper = mid

    return upper



from typing import Callable, MutableMapping, TypeVar, Union, cast
import inspect

from sarus_data_spec.protobuf import type_name
from sarus_data_spec.protobuf.typing import Protobuf
import sarus_data_spec.typing as st

ReferrableBuilder = Callable[[Protobuf, bool], st.HasProtobuf]
BaseBuilder = Callable[[Protobuf], st.HasProtobuf]

Builder = Union[ReferrableBuilder, BaseBuilder]


class Factory(st.Factory):
    """Can produce objects from protobuf messages"""

    def __init__(self) -> None:
        self.type_name_create: MutableMapping[str, Builder] = {}

    def register(self, name: str, create: Builder) -> None:
        self.type_name_create[name] = create

    M = TypeVar('M', bound=Protobuf)

    def create(self, message: M, store: bool = True) -> st.HasProtobuf[M]:
        """Instantiate a wrapping class from a protobuf message.

        Factory functions (builders) may optionally take a `store` parameter.
        We check dynamically if the registered builder accepts the `store`
        parameter and pass it the `store` parameter accordingly.
        """
        builder = self.type_name_create[type_name(message)]
        args = list(inspect.signature(builder).parameters.keys())
        if len(args) == 2 and "store" in args:
            referrable_builder = cast(ReferrableBuilder, builder)
            return referrable_builder(message, store)
        elif len(args) == 1:
            base_builder = cast(BaseBuilder, builder)
            return base_builder(message)
        else:
            raise ValueError(f"Unknown factory function signature {args}")



from __future__ import annotations

from typing import Type

from sarus_data_spec.base import Base
import sarus_data_spec.protobuf as sp
import sarus_data_spec.typing as st


class Predicate(Base[sp.Predicate]):
    """A python class to describe types"""

    def prototype(self) -> Type[sp.Predicate]:
        """Return the type of the underlying protobuf."""
        return sp.Predicate

    # A bunch of operators
    def __or__(self, predicate: st.Predicate) -> Predicate:
        """Union operator"""
        if self._protobuf.HasField('union'):
            result = Predicate(self._protobuf)
            if predicate.protobuf().HasField('union'):
                result._protobuf.union.predicates.extend(
                    predicate.protobuf().union.predicates
                )
            else:
                result._protobuf.union.predicates.append(predicate.protobuf())
            return result
        else:
            if predicate.protobuf().HasField('union'):
                result = Predicate(predicate.protobuf())
                result._protobuf.union.predicates.insert(0, self._protobuf)
                return result
            else:
                return union(self, predicate)

    def __and__(self, predicate: st.Predicate) -> Predicate:
        """Inter operator"""
        if self._protobuf.HasField('inter'):
            result = Predicate(self._protobuf)
            if predicate.protobuf().HasField('inter'):
                result._protobuf.inter.predicates.extend(
                    predicate.protobuf().inter.predicates
                )
            else:
                result._protobuf.inter.predicates.append(predicate.protobuf())
            return result
        else:
            if predicate.protobuf().HasField('inter'):
                result = Predicate(predicate.protobuf())
                result._protobuf.inter.predicates.insert(0, self._protobuf)
                return result
            else:
                return inter(self, predicate)

    def __invert__(self) -> Predicate:
        """Complement"""
        if self._protobuf.HasField('comp'):
            return Predicate(self._protobuf.comp.predicate)
        else:
            return comp(self)


def inter(*predicates: st.Predicate) -> Predicate:
    return Predicate(
        sp.Predicate(
            inter=sp.Predicate.Inter(
                predicates=(p.protobuf() for p in predicates)
            )
        )
    )


def union(*predicates: st.Predicate) -> Predicate:
    return Predicate(
        sp.Predicate(
            union=sp.Predicate.Union(
                predicates=(p.protobuf() for p in predicates)
            )
        )
    )


def comp(predicate: st.Predicate) -> Predicate:
    return Predicate(
        sp.Predicate(comp=sp.Predicate.Comp(predicate=predicate.protobuf()))
    )


def simple(operator: str, value: str) -> Predicate:
    return Predicate(
        sp.Predicate(
            simple=sp.Predicate.Simple(operator=operator, value=value)
        )
    )



from typing import Any, Literal, Optional, Union

import numpy as np

from sarus_data_spec.dataspec_validator.signature import (
    SarusParameter,
    SarusSignature,
    SarusSignatureValue,
)

from ..external_op import ExternalOpImplementation

try:
    from sklearn import svm
except ModuleNotFoundError:
    pass  # error message in typing.py


class sk_svc(ExternalOpImplementation):
    _transform_id = "sklearn.SK_SVC"
    _signature = SarusSignature(
        SarusParameter(
            name="C",
            annotation=float,
            default=1.0,
        ),
        SarusParameter(
            name="kernel",
            annotation=Union[
                Literal['linear', 'poly', 'rbf', 'sigmoid', 'precomputed'],
                callable,
            ],
            default="rbf",
        ),
        SarusParameter(
            name="degree",
            annotation=int,
            default=3,
        ),
        SarusParameter(
            name="gamma",
            annotation=Union[Literal['scale', 'auto'], float],
            default="scale",
        ),
        SarusParameter(
            name="coef0",
            annotation=float,
            default=0.0,
        ),
        SarusParameter(
            name="shrinking",
            annotation=bool,
            default=True,
        ),
        SarusParameter(
            name="probability",
            annotation=bool,
            default=False,
        ),
        SarusParameter(
            name="tol",
            annotation=float,
            default=1e-3,
        ),
        SarusParameter(
            name="cache_size",
            annotation=float,
            default=200,
        ),
        SarusParameter(
            name="class_weight",
            annotation=Optional[Union[Literal["balanced"], dict]],
            default=None,
        ),
        SarusParameter(
            name="verbose",
            annotation=bool,
            default=False,
        ),
        SarusParameter(
            name="max_iter",
            annotation=int,
            default=-1,
        ),
        SarusParameter(
            name="decision_function_shape",
            annotation=Literal['ovo', 'ovr'],
            default="ovr",
        ),
        SarusParameter(
            name="break_ties",
            annotation=bool,
            default=False,
        ),
        SarusParameter(
            name="random_state",
            annotation=Optional[Union[int, np.random.RandomState]],
            default=None,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        kwargs = signature.collect_kwargs()
        return svm.SVC(**kwargs)



from __future__ import annotations

from uuid import UUID
import asyncio
import hashlib
import logging
import os
import time
import typing as t
import warnings

import pandas as pd
import pyarrow as pa

from sarus_data_spec.arrow.admin_utils import async_to_arrow_extract_data_only
from sarus_data_spec.arrow.array import convert_record_batch
from sarus_data_spec.arrow.conversion import async_cast_arrow_batches
from sarus_data_spec.attribute import attach_properties
from sarus_data_spec.constants import (
    BIG_DATA_TASK,
    BIG_DATA_THRESHOLD,
    DATASET_N_BYTES,
    DATASET_N_LINES,
    IS_BIG_DATA,
    IS_REMOTE,
    THRESHOLD_TYPE,
)
from sarus_data_spec.dataspec_rewriter.base import BaseDataspecRewriter
from sarus_data_spec.dataspec_validator.base import BaseDataspecValidator
from sarus_data_spec.manager.async_utils import sync, sync_iterator
from sarus_data_spec.manager.ops.processor.external.external_op import (  # noqa: E501
    async_compute_external_value,
)
from sarus_data_spec.manager.ops.processor.routing import get_implementation
from sarus_data_spec.manager.typing import Computation
from sarus_data_spec.protobuf.utilities import copy
from sarus_data_spec.protobuf.utilities import json as utilities_json
from sarus_data_spec.scalar import Scalar
from sarus_data_spec.schema import Schema
from sarus_data_spec.status import error_aggregation
from sarus_data_spec.transform import transform_id
import sarus_data_spec.dataspec_rewriter.typing as sdrt
import sarus_data_spec.dataspec_validator.typing as sdvt
import sarus_data_spec.manager.typing as manager_typing
import sarus_data_spec.protobuf as sp
import sarus_data_spec.status as stt
import sarus_data_spec.storage.typing as storage_typing
import sarus_data_spec.typing as st

try:
    from sarus_data_spec.bounds import Bounds
    from sarus_data_spec.links import Links
    from sarus_data_spec.marginals import Marginals
    from sarus_data_spec.multiplicity import Multiplicity
    from sarus_data_spec.size import Size

except ModuleNotFoundError:
    pass
try:
    import tensorflow as tf

    from sarus_data_spec.manager.ops.tensorflow.features import (
        deserialize,
        flatten,
        nest,
        serialize,
        to_internal_signature,
    )
    from sarus_data_spec.manager.ops.tensorflow.tensorflow_visitor import (  # noqa: E501
        convert_tensorflow,
    )
except ModuleNotFoundError:
    pass  # error message printed from typing.py


logger = logging.getLogger(__name__)

BATCH_SIZE = 10000


class Base(manager_typing.Manager):
    """Provide the dataset functionalities."""

    def __init__(
        self, storage: storage_typing.Storage, protobuf: sp.Manager
    ) -> None:
        self._protobuf: sp.Manager = copy(protobuf)
        self._freeze()
        self._storage = storage
        self.storage().store(self)
        self._parquet_dir = os.path.expanduser('/tmp/sarus_dataset/')
        os.makedirs(self.parquet_dir(), exist_ok=True)
        self._dataspec_rewriter = BaseDataspecRewriter(storage=storage)
        self._dataspec_validator = BaseDataspecValidator(storage=storage)

        # To define in subclasses
        self.to_arrow_computation: Computation[t.AsyncIterator[pa.RecordBatch]]
        self.to_parquet_computation: Computation[str]
        self.cache_scalar_computation: Computation[t.Tuple[str, str]]
        self.value_computation: Computation[t.Any]
        self.schema_computation: Computation[st.Schema]

    # --------PROTOBUF METHODS---------
    def protobuf(self) -> sp.Manager:
        return copy(self._protobuf)

    def prototype(self) -> t.Type[sp.Manager]:
        return sp.Manager

    def type_name(self) -> str:
        return sp.type_name(self._protobuf)

    def __repr__(self) -> str:
        return utilities_json(self._protobuf)

    def __getitem__(self, key: str) -> str:
        return t.cast(str, self._protobuf.properties[key])

    def properties(self) -> t.Mapping[str, str]:
        return self.protobuf().properties

    def _checksum(self) -> bytes:
        """Compute an md5 checksum"""
        md5 = hashlib.md5()
        md5.update(sp.serialize(self._protobuf))
        return md5.digest()

    def _freeze(self) -> None:
        self._protobuf.uuid = ''
        self._frozen_checksum = self._checksum()
        self._protobuf.uuid = UUID(bytes=self._frozen_checksum).hex

    def _frozen(self) -> bool:
        uuid = self._protobuf.uuid
        self._protobuf.uuid = ''
        result = (self._checksum() == self._frozen_checksum) and (
            uuid == UUID(bytes=self._frozen_checksum).hex
        )
        self._protobuf.uuid = uuid
        return result

    def uuid(self) -> str:
        return self._protobuf.uuid

    def referring(
        self, type_name: t.Optional[str] = None
    ) -> t.Collection[st.Referring]:
        return self.storage().referring(self, type_name=type_name)

    def storage(self) -> storage_typing.Storage:
        return self._storage

    # -------DATASPEC RELATED------

    def attribute(
        self, name: str, dataspec: st.DataSpec
    ) -> t.Optional[st.Attribute]:
        attributes = t.cast(
            t.Set[st.Attribute],
            self.storage().referring(
                dataspec, type_name=sp.type_name(sp.Attribute)
            ),
        )
        filtered_attributes = [
            element for element in attributes if element.name() == name
        ]

        if len(filtered_attributes) == 0:
            return None

        if len(filtered_attributes) == 1:
            return t.cast(st.Attribute, filtered_attributes[0])

        raise ValueError('There are two attributes with the same name')

    def attributes(
        self, name: str, dataspec: st.DataSpec
    ) -> t.List[st.Attribute]:
        attributes = t.cast(
            t.Set[st.Attribute],
            self.storage().referring(
                dataspec, type_name=sp.type_name(sp.Attribute)
            ),
        )
        return [element for element in attributes if element.name() == name]

    def infer_output_type(
        self,
        transform: st.Transform,
        *arguments: t.Union[st.DataSpec, st.Transform],
        **named_arguments: t.Union[st.DataSpec, st.Transform],
    ) -> t.Tuple[str, t.Callable[[st.DataSpec], None]]:
        """Infer the transform output type : minimal type inference."""

        def attach_nothing(ds: st.DataSpec) -> None:
            return

        # Some internal transforms return a Scalar
        if transform.spec() in [
            'protected_paths',
            'automatic_user_settings',
            'public_paths',
            'automatic_budget',
            'attribute_budget',
            'sd_budget',
            'sampling_ratios',
            'derive_seed',
            'relationship_spec',
            'validated_user_type',
        ]:
            return sp.type_name(sp.Scalar), attach_nothing

        # Other internal transforms return a Dataset
        if not transform.is_external():
            return sp.type_name(sp.Dataset), attach_nothing

        # Compute the mock value for external transforms
        mock_value = self.mock_value(transform, *arguments, **named_arguments)

        # Infer output types
        dataspec_type = (
            sp.type_name(sp.Dataset)
            if type(mock_value) in t.get_args(st.DatasetCastable)
            else sp.type_name(sp.Scalar)
        )

        # Return the output type
        return dataspec_type, attach_nothing

    def is_big_data(self, dataset: st.Dataset) -> bool:
        status = self.status(dataset, task_name=BIG_DATA_TASK)

        if status is not None:
            # check if big_data_present
            big_data_task = status.task(BIG_DATA_TASK)
            # if yes:return answer
            if (big_data_task is not None) and (
                big_data_task.stage() == 'ready'
            ):
                return big_data_task.properties()[IS_BIG_DATA] == str(True)

        if dataset.is_source():
            raise NotImplementedError(
                'Found source dataset without any big data status'
            )
        else:
            if dataset.transform().is_external():
                return False
            parents, named_parents = dataset.parents()
            parents_list = [
                parent for parent in parents if isinstance(parent, st.DataSpec)
            ]
            parents_dict = {
                name: parent
                for name, parent in named_parents.items()
                if isinstance(parent, st.DataSpec)
            }
            ds_args = [
                element
                for element in parents_list
                if element.type_name() == sp.type_name(sp.Dataset)
            ]
            for element in parents_dict.values():
                if element.type_name() == sp.type_name(sp.Dataset):
                    ds_args.append(element)
            # parents_list.extend(list(parents_dict.values()))
            # parents_list=t.cast(t.Sequence[st.Dataset],parents_list)
            if len(ds_args) > 1:
                raise NotImplementedError(
                    'transforms with many dataspecs not supported yet'
                )
            is_parent_big_data = self.is_big_data(
                ds_args[0]  # type:ignore
            )

            if not is_parent_big_data:
                # write status it is not big data
                stt.ready(
                    dataset,
                    task=BIG_DATA_TASK,
                    properties={
                        IS_BIG_DATA: str(False)
                    },  # we do not need to add more info because a
                    # non big_data dataset cannot become big_data
                )
                return False
            else:
                # check that the transform does not change
                # the big data status
                (
                    is_big_data,
                    number_lines,
                    number_bytes,
                    threshold_kind,
                ) = check_transform_big_data(
                    dataset.transform(),
                    parents_list[0],  # type:ignore
                    parents_dict,
                )
                big_data_threshold = int(
                    stt.last_status(
                        dataspec=parents_list[0], task=BIG_DATA_TASK
                    )  # type:ignore
                    .task(BIG_DATA_TASK)
                    .properties()[BIG_DATA_THRESHOLD]
                )
                # write status
                stt.ready(
                    dataset,
                    task=BIG_DATA_TASK,
                    properties={
                        IS_BIG_DATA: str(is_big_data),
                        BIG_DATA_THRESHOLD: str(big_data_threshold),
                        DATASET_N_LINES: str(number_lines),
                        DATASET_N_BYTES: str(number_bytes),
                        THRESHOLD_TYPE: threshold_kind,
                    },
                )
                return is_big_data

    def status(
        self, dataspec: st.DataSpec, task_name: t.Optional[str] = None
    ) -> t.Optional[st.Status]:
        return stt.last_status(
            dataspec=dataspec,
            manager=self,
            task=task_name,
        )

    def foreign_keys(self, dataset: st.Dataset) -> t.Dict[st.Path, st.Path]:
        return t.cast(
            t.Dict[st.Path, st.Path], sync(self.async_foreign_keys(dataset))
        )

    def primary_keys(self, dataset: st.Dataset) -> t.List[st.Path]:
        return t.cast(t.List[st.Path], sync(self.async_primary_keys(dataset)))

    async def async_primary_keys(self, dataset: st.Dataset) -> t.List[st.Path]:
        raise NotImplementedError

    async def async_foreign_keys(
        self, dataset: st.Dataset
    ) -> t.Dict[st.Path, st.Path]:
        raise NotImplementedError

    # -------CACHING POLICY/UTILS ----

    def is_cached(self, dataspec: st.DataSpec) -> bool:
        raise NotImplementedError

    def is_pushed_to_sql(self, dataspec: st.DataSpec) -> bool:
        raise NotImplementedError

    def parquet_dir(self) -> str:
        return self._parquet_dir

    def sql_pushing_schema_prefix(self, dataset: st.Dataset) -> str:
        raise NotImplementedError

    def computation_timeout(self, dataspec: st.DataSpec) -> int:
        raise NotImplementedError

    def computation_max_delay(self, dataspec: st.DataSpec) -> int:
        raise NotImplementedError

    # ------SYNC COMPUTATIONS METHODS--------------

    def to(
        self, dataset: st.Dataset, kind: t.Type, drop_admin: bool = True
    ) -> st.DatasetCastable:
        return sync(self.async_to(dataset, kind, drop_admin))

    def bounds(self, dataset: st.Dataset) -> st.Bounds:
        return t.cast(Bounds, sync(self.async_bounds(dataset)))

    def cache_scalar(self, scalar: st.Scalar) -> None:
        sync(self.async_cache_scalar(scalar=scalar))

    def links(self, dataset: st.Dataset) -> st.Links:
        return t.cast(Links, sync(self.async_links(dataset)))

    def marginals(self, dataset: st.Dataset) -> st.Marginals:
        return t.cast(Marginals, sync(self.async_marginals(dataset)))

    def prepare(self, dataspec: st.DataSpec) -> None:
        """Make sure a Dataspec is ready."""
        sync(self.async_prepare(dataspec))

    def schema(self, dataset: st.Dataset) -> Schema:
        return t.cast(Schema, sync(self.async_schema(dataset=dataset)))

    def size(self, dataset: st.Dataset) -> st.Size:
        return t.cast(Size, sync(self.async_size(dataset)))

    def multiplicity(self, dataset: st.Dataset) -> st.Multiplicity:
        return t.cast(Multiplicity, sync(self.async_multiplicity(dataset)))

    def sql(
        self,
        dataset: st.Dataset,
        query: t.Union[str, t.Dict[str, t.Any]],
        dialect: t.Optional[st.SQLDialect] = None,
        batch_size: int = 10000,
    ) -> t.Iterator[pa.RecordBatch]:
        return sync_iterator(
            self.async_sql(dataset, query, dialect, batch_size)
        )

    def sql_prepare(self, dataset: st.Dataset) -> None:
        """SQL prepare dataset."""
        sync(self.async_sql_prepare(dataset))

    def to_arrow(
        self, dataset: st.Dataset, batch_size: int
    ) -> t.Iterator[pa.RecordBatch]:
        return sync_iterator(
            self.async_to_arrow(dataset=dataset, batch_size=batch_size)
        )

    def to_pandas(self, dataset: st.Dataset) -> pd.DataFrame:
        return sync(self.async_to_pandas(dataset=dataset))

    def to_parquet(self, dataset: st.Dataset) -> None:
        sync(self.async_to_parquet(dataset=dataset))

    def to_sql(self, dataset: st.Dataset) -> None:
        sync(self.async_to_sql(dataset=dataset))

    def to_tensorflow(self, dataset: st.Dataset) -> tf.data.Dataset:
        return sync(self.async_to_tensorflow(dataset=dataset))

    def value(self, scalar: st.Scalar) -> st.DataSpecValue:
        return sync(self.async_value(scalar=scalar))

    def composed_callable(
        self, transform: st.Transform
    ) -> t.Callable[..., t.Any]:
        implementation = get_implementation(transform.transform_to_apply())
        return implementation.callable(transform)

    # ------ASYNC COMPUTATIONS--------
    async def async_to(
        self, dataset: st.Dataset, kind: t.Type, drop_admin: bool = True
    ) -> st.DatasetCastable:
        """Convert a Dataset's to a Python type."""
        if drop_admin:
            batches_async_it = await async_to_arrow_extract_data_only(
                dataset=dataset, batch_size=BATCH_SIZE
            )
        else:
            if kind not in [pd.DataFrame, t.Iterator[pa.RecordBatch]]:
                raise TypeError(
                    f"The target type {kind} is not compatible"
                    " with the protection."
                )
            batches_async_it = await dataset.async_to_arrow(
                batch_size=BATCH_SIZE
            )

        return await async_cast_arrow_batches(batches_async_it, kind)

    async def async_bounds(self, dataset: st.Dataset) -> st.Bounds:
        raise NotImplementedError

    async def async_cache_scalar(self, scalar: st.Scalar) -> None:
        await self.cache_scalar_computation.complete_task(dataspec=scalar)

    async def async_prepare(self, dataspec: st.DataSpec) -> None:
        """Make sure a Dataspec is ready asynchronously."""
        computation = self.dataspec_computation(dataspec)
        await computation.complete_task(dataspec)

    async def async_prepare_parents(self, dataspec: st.DataSpec) -> None:
        """Prepare all the parents of a Dataspec."""
        args, kwargs = dataspec.parents()
        parents = list(args) + list(kwargs.values())
        coros = [
            self.async_prepare(parent)
            for parent in parents
            if isinstance(parent, st.DataSpec)
        ]
        # here, if many parents potentially fail, we want to be sure
        # that all of them do it, and not only the first one (to
        # modify all statuses accordingly).
        # After, we only raise the first exception for the child.

        results = await asyncio.gather(*coros, return_exceptions=True)
        exceptions = [
            element for element in results if isinstance(element, Exception)
        ]
        if len(exceptions) == 0:
            return
        raise error_aggregation(exceptions)

    async def async_links(self, dataset: st.Dataset) -> t.Any:
        raise NotImplementedError

    async def async_marginals(self, dataset: st.Dataset) -> st.Marginals:
        raise NotImplementedError

    async def async_schema(self, dataset: st.Dataset) -> st.Schema:
        return await self.schema_computation.task_result(dataspec=dataset)

    async def async_size(self, dataset: st.Dataset) -> st.Size:
        raise NotImplementedError

    async def async_multiplicity(self, dataset: st.Dataset) -> st.Multiplicity:
        raise NotImplementedError

    async def async_to_arrow(
        self, dataset: st.Dataset, batch_size: int
    ) -> t.AsyncIterator[pa.RecordBatch]:
        """Reads asynchronous iterator of datast batches"""
        computation = t.cast(
            Computation[t.AsyncIterator[pa.RecordBatch]],
            self.dataspec_computation(dataset),
        )
        return await computation.task_result(
            dataspec=dataset, batch_size=batch_size
        )

    async def async_to_parquet(self, dataset: st.Dataset) -> None:
        await self.to_parquet_computation.complete_task(dataspec=dataset)

    async def async_sql(
        self,
        dataset: st.Dataset,
        query: t.Union[str, t.Dict[str, t.Any]],
        dialect: t.Optional[st.SQLDialect] = None,
        batch_size: int = 10000,
        result_type: t.Optional[st.Type] = None,
    ) -> t.AsyncIterator[pa.RecordBatch]:
        computation = t.cast(
            Computation[t.AsyncIterator[pa.RecordBatch]],
            self.sql_computation(),
        )
        return await computation.task_result(
            dataset,
            query=query,
            dialect=dialect,
            batch_size=batch_size,
            result_type=result_type,
        )

    async def async_sql_prepare(self, dataset: st.Dataset) -> None:
        """SQL prepare the dataset synchronously"""
        computation = t.cast(
            Computation[t.AsyncIterator[pa.RecordBatch]],
            self.sql_computation(),
        )
        await computation.complete_task(dataset)

    async def async_sql_prepare_parents(self, dataset: st.Dataset) -> None:
        """SQL prepare all the parents of a dataset: calling async_sql_prepare
        for dataset parents and async_prepare for scalar parents.
        """
        args, kwargs = dataset.parents()
        parents = list(args) + list(kwargs.values())
        coros = [
            self.async_sql_prepare(t.cast(st.Dataset, parent))
            if dataset.prototype() == sp.Dataset
            else self.async_prepare(parent)
            for parent in parents
            if isinstance(parent, st.DataSpec)
        ]
        # here, if many parents potentially fail, we want to be sure
        # that all of them do it, and not only the first one (to
        # modify all statuses accordingly).
        # After, we only raise the first exception for the child.

        results = await asyncio.gather(*coros, return_exceptions=True)
        exceptions = [
            element for element in results if isinstance(element, Exception)
        ]
        if len(exceptions) == 0:
            return
        raise error_aggregation(exceptions)

    async def async_to_pandas(self, dataset: st.Dataset) -> pd.DataFrame:
        return await self.async_to(
            dataset=dataset, kind=pd.DataFrame, drop_admin=False
        )

    async def async_value(self, scalar: st.Scalar) -> t.Any:
        """Reads asynchronously value of a scalar."""
        computation = self.dataspec_computation(scalar)
        return await computation.task_result(dataspec=scalar)

    # -------- COMPILATION---------

    def dataspec_rewriter(self) -> sdrt.DataspecRewriter:
        return self._dataspec_rewriter

    def dataspec_validator(self) -> sdvt.DataspecValidator:
        return self._dataspec_validator

    def set_remote(self, dataspec: st.DataSpec) -> None:
        """Add an Attribute to tag the DataSpec as remotely fetched."""
        attach_properties(
            dataspec,
            properties={
                # TODO deprecated in SDS >=2.0.0 -> remove property
                "is_remote": str(True)
            },
            name=IS_REMOTE,
        )

    def is_remote(self, dataspec: st.DataSpec) -> bool:
        """Is the dataspec a remotely defined dataset."""
        is_remote_att = self.attribute(IS_REMOTE, dataspec)
        return is_remote_att is not None

    def dataspec_computation(
        self, dataspec: st.DataSpec
    ) -> Computation[t.AsyncIterator[pa.RecordBatch]]:
        """Return the Computation for getting the dataspec's value"""
        raise NotImplementedError

    def sql_computation(self) -> Computation[t.AsyncIterator[pa.RecordBatch]]:
        """Returns the SQL Computation of the manager. The aim of the method is
        is to reduce code duplication between API and Worker manager
        """
        raise NotImplementedError

    async def async_to_tensorflow(
        self, dataset: st.Dataset
    ) -> tf.data.Dataset:
        root_dir = os.path.join(
            self.parquet_dir(), "tfrecords", dataset.uuid()
        )
        schema_type = (await self.async_schema(dataset)).type()
        signature = to_internal_signature(schema_type)

        if not os.path.exists(root_dir):
            # the dataset is cached first
            os.makedirs(root_dir)

            flattener = flatten(signature)
            serializer = serialize(signature)
            i = 0
            batches_async_it = await self.async_to_arrow(
                dataset=dataset, batch_size=BATCH_SIZE
            )
            async for batch in batches_async_it:
                filename = os.path.join(root_dir, f"batch_{i}.tfrecord")
                i += 1
                await write_tf_batch(
                    filename, batch, schema_type, flattener, serializer
                )

        # reading from cache
        glob = os.path.join(root_dir, "*.tfrecord")
        filenames = tf.data.Dataset.list_files(glob, shuffle=False)
        deserializer = deserialize(signature)
        nester = nest(signature)
        return tf.data.TFRecordDataset(filenames).map(deserializer).map(nester)

    def mock_value(
        self,
        transform: st.Transform,
        *arguments: t.Union[st.DataSpec, st.Transform],
        **named_arguments: t.Union[st.DataSpec, st.Transform],
    ) -> t.Any:
        """Compute the mock value of an external transform applied on
        Dataspecs.
        """
        start = time.perf_counter()
        assert transform.is_external()
        mock_args = [
            arg.variant(st.ConstraintKind.MOCK)
            if isinstance(arg, st.DataSpec)
            else arg  # `arg` can be a composed Transform
            for arg in arguments
        ]
        named_mock_args = {
            name: arg.variant(st.ConstraintKind.MOCK)
            if isinstance(arg, st.DataSpec)
            else arg  # `arg` can be a composed Transform
            for name, arg in named_arguments.items()
        }

        if any([ds is None for ds in mock_args]) or any(
            [ds is None for ds in named_mock_args.values()]
        ):
            raise ValueError(
                "Cannot infer the output type of the external "
                "transform because one of the parents has a None MOCK."
            )

        typed_mock_args = [t.cast(st.DataSpec, ds) for ds in mock_args]
        typed_named_mock_args = {
            name: t.cast(st.DataSpec, ds)
            for name, ds in named_mock_args.items()
        }

        try:
            mock_value = sync(
                async_compute_external_value(
                    transform, *typed_mock_args, **typed_named_mock_args
                )
            )
        except Exception as e:
            raise e

        end = time.perf_counter()
        logger.debug(
            f"MOCK VALUE {transform_id(transform)} ({end-start:.2f}s)"
        )

        return mock_value

    async def async_value_op(self, scalar: st.Scalar) -> t.Any:
        raise NotImplementedError

    async def async_to_arrow_op(
        self, dataset: st.Dataset, batch_size: int
    ) -> t.AsyncIterator[pa.RecordBatch]:
        raise NotImplementedError

    async def async_schema_op(self, dataset: st.Dataset) -> st.Schema:
        raise NotImplementedError


def check_transform_big_data(
    transform: st.Transform,
    parent_dataset: st.Dataset,
    parents_dict: t.Mapping[str, st.DataSpec],
) -> t.Tuple[bool, int, int, str]:
    """This methods return true if the dataset transformed
    is big_data and False otherwise. This method is called when the parent
    is big_data so if the transform does not
    affect the size, it should return True
    """
    status = stt.last_status(parent_dataset, task=BIG_DATA_TASK)
    assert status
    big_data_threshold = int(
        status.task(BIG_DATA_TASK).properties()[  # type:ignore
            BIG_DATA_THRESHOLD
        ]
    )
    threshold_kind = status.task(BIG_DATA_TASK).properties()[  # type:ignore
        THRESHOLD_TYPE
    ]

    parent_n_lines_str = status.task(
        BIG_DATA_TASK
    ).properties()[  # type:ignore
        DATASET_N_LINES
    ]
    if parent_n_lines_str == '':
        parent_n_lines = 0
    else:
        parent_n_lines = int(parent_n_lines_str)

    parent_bytes_str = status.task(BIG_DATA_TASK).properties()[  # type:ignore
        DATASET_N_BYTES
    ]
    if parent_bytes_str == '':
        parent_bytes = 0
    else:
        parent_bytes = int(parent_bytes_str)

    transform_name = transform.name()
    if transform_name in ('Sample', 'DifferentiatedSample'):
        transform_type = transform.protobuf().spec.WhichOneof('spec')
        assert transform_type
        if getattr(transform.protobuf().spec, transform_type).HasField(
            'fraction'
        ):
            fraction = getattr(
                transform.protobuf().spec, transform_type
            ).fraction
            new_bytes = int(fraction * parent_bytes)
            n_lines = int(fraction * parent_n_lines)

            if threshold_kind == DATASET_N_BYTES:
                return (
                    new_bytes > big_data_threshold,
                    n_lines,
                    new_bytes,
                    threshold_kind,
                )
        else:
            n_lines = getattr(transform.protobuf().spec, transform_type).size
            new_bytes = int(n_lines * parent_bytes / parent_n_lines)

            if threshold_kind == DATASET_N_BYTES:
                big_data_threshold = int(1e5)

        threshold_kind = DATASET_N_LINES

        return n_lines > big_data_threshold, n_lines, new_bytes, threshold_kind

    elif transform_name == 'filter':
        # TODO: we need to save the real sizes of a dataspec in the statuses
        # so that we can check what happens here
        raise NotImplementedError(
            ' Big data status tot implemented yet for filtering'
        )

    elif transform_name == 'Synthetic data':
        # here we should leverage the sampling ratio just to get the size,
        # in any case, synthetic data is never big data
        sampling_ratios = t.cast(
            t.Mapping, t.cast(Scalar, parents_dict['sampling_ratios']).value()
        )
        synthetic_size = 0
        for table_path, sampling_ratio in zip(
            parent_dataset.schema().tables(), sampling_ratios.values()
        ):
            size = parent_dataset.size()
            assert size
            stat = size.statistics().nodes_statistics(table_path)[0]
            synthetic_size += int(stat.size() * sampling_ratio)
        threshold_kind = DATASET_N_LINES
        return (
            False,
            synthetic_size,
            int(synthetic_size * parent_bytes / parent_n_lines),
            threshold_kind,
        )
    elif transform_name == "select_sql":
        # TODO https://gitlab.com/sarus-tech/sarus-data-spec/-/issues/207
        return (False, -1, -1, threshold_kind)

    else:
        # other transforms do not affect size
        if transform_name == 'user_settings':
            warnings.warn(
                'user_settings transform considered to' 'not affect size'
            )
        return True, parent_n_lines, parent_bytes, threshold_kind


async def write_tf_batch(
    filename: str,
    batch: pa.RecordBatch,
    schema_type: st.Type,
    flattener: t.Callable,
    serializer: t.Callable,
) -> None:
    with tf.io.TFRecordWriter(filename) as writer:
        batch = convert_tensorflow(
            convert_record_batch(record_batch=batch, _type=schema_type),
            schema_type,
        )
        batch = tf.data.Dataset.from_tensor_slices(batch).map(flattener)
        for row in batch:
            as_bytes = serializer(row)
            writer.write(as_bytes)



"""
@generated by mypy-protobuf.  Do not edit manually!
isort:skip_file
"""
import builtins
import google.protobuf.descriptor
import google.protobuf.internal.containers
import google.protobuf.message
import typing
import typing_extensions

DESCRIPTOR: google.protobuf.descriptor.FileDescriptor = ...

class Status(google.protobuf.message.Message):
    """Current form of the dataset"""
    DESCRIPTOR: google.protobuf.descriptor.Descriptor = ...
    class TaskStagesEntry(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor = ...
        KEY_FIELD_NUMBER: builtins.int
        VALUE_FIELD_NUMBER: builtins.int
        key: typing.Text = ...
        @property
        def value(self) -> global___Status.Stage: ...
        def __init__(self,
            *,
            key : typing.Text = ...,
            value : typing.Optional[global___Status.Stage] = ...,
            ) -> None: ...
        def HasField(self, field_name: typing_extensions.Literal["value",b"value"]) -> builtins.bool: ...
        def ClearField(self, field_name: typing_extensions.Literal["key",b"key","value",b"value"]) -> None: ...

    class PropertiesEntry(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor = ...
        KEY_FIELD_NUMBER: builtins.int
        VALUE_FIELD_NUMBER: builtins.int
        key: typing.Text = ...
        value: typing.Text = ...
        def __init__(self,
            *,
            key : typing.Text = ...,
            value : typing.Text = ...,
            ) -> None: ...
        def ClearField(self, field_name: typing_extensions.Literal["key",b"key","value",b"value"]) -> None: ...

    class Stage(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor = ...
        class PropertiesEntry(google.protobuf.message.Message):
            DESCRIPTOR: google.protobuf.descriptor.Descriptor = ...
            KEY_FIELD_NUMBER: builtins.int
            VALUE_FIELD_NUMBER: builtins.int
            key: typing.Text = ...
            value: typing.Text = ...
            def __init__(self,
                *,
                key : typing.Text = ...,
                value : typing.Text = ...,
                ) -> None: ...
            def ClearField(self, field_name: typing_extensions.Literal["key",b"key","value",b"value"]) -> None: ...

        class Pending(google.protobuf.message.Message):
            DESCRIPTOR: google.protobuf.descriptor.Descriptor = ...
            def __init__(self,
                ) -> None: ...

        class Processing(google.protobuf.message.Message):
            DESCRIPTOR: google.protobuf.descriptor.Descriptor = ...
            def __init__(self,
                ) -> None: ...

        class Ready(google.protobuf.message.Message):
            DESCRIPTOR: google.protobuf.descriptor.Descriptor = ...
            def __init__(self,
                ) -> None: ...

        class Error(google.protobuf.message.Message):
            DESCRIPTOR: google.protobuf.descriptor.Descriptor = ...
            def __init__(self,
                ) -> None: ...

        PENDING_FIELD_NUMBER: builtins.int
        PROCESSING_FIELD_NUMBER: builtins.int
        READY_FIELD_NUMBER: builtins.int
        ERROR_FIELD_NUMBER: builtins.int
        PROPERTIES_FIELD_NUMBER: builtins.int
        @property
        def pending(self) -> global___Status.Stage.Pending: ...
        @property
        def processing(self) -> global___Status.Stage.Processing: ...
        @property
        def ready(self) -> global___Status.Stage.Ready: ...
        @property
        def error(self) -> global___Status.Stage.Error: ...
        @property
        def properties(self) -> google.protobuf.internal.containers.ScalarMap[typing.Text, typing.Text]: ...
        def __init__(self,
            *,
            pending : typing.Optional[global___Status.Stage.Pending] = ...,
            processing : typing.Optional[global___Status.Stage.Processing] = ...,
            ready : typing.Optional[global___Status.Stage.Ready] = ...,
            error : typing.Optional[global___Status.Stage.Error] = ...,
            properties : typing.Optional[typing.Mapping[typing.Text, typing.Text]] = ...,
            ) -> None: ...
        def HasField(self, field_name: typing_extensions.Literal["error",b"error","pending",b"pending","processing",b"processing","ready",b"ready","stage",b"stage"]) -> builtins.bool: ...
        def ClearField(self, field_name: typing_extensions.Literal["error",b"error","pending",b"pending","processing",b"processing","properties",b"properties","ready",b"ready","stage",b"stage"]) -> None: ...
        def WhichOneof(self, oneof_group: typing_extensions.Literal["stage",b"stage"]) -> typing.Optional[typing_extensions.Literal["pending","processing","ready","error"]]: ...

    UUID_FIELD_NUMBER: builtins.int
    DATASPEC_FIELD_NUMBER: builtins.int
    MANAGER_FIELD_NUMBER: builtins.int
    DATETIME_FIELD_NUMBER: builtins.int
    TASK_STAGES_FIELD_NUMBER: builtins.int
    PROPERTIES_FIELD_NUMBER: builtins.int
    uuid: typing.Text = ...
    dataspec: typing.Text = ...
    manager: typing.Text = ...
    datetime: typing.Text = ...
    """ISO 8601 datetime"""

    @property
    def task_stages(self) -> google.protobuf.internal.containers.MessageMap[typing.Text, global___Status.Stage]:
        """Various tasks handled by the manager"""
        pass
    @property
    def properties(self) -> google.protobuf.internal.containers.ScalarMap[typing.Text, typing.Text]:
        """Other properties"""
        pass
    def __init__(self,
        *,
        uuid : typing.Text = ...,
        dataspec : typing.Text = ...,
        manager : typing.Text = ...,
        datetime : typing.Text = ...,
        task_stages : typing.Optional[typing.Mapping[typing.Text, global___Status.Stage]] = ...,
        properties : typing.Optional[typing.Mapping[typing.Text, typing.Text]] = ...,
        ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["dataspec",b"dataspec","datetime",b"datetime","manager",b"manager","properties",b"properties","task_stages",b"task_stages","uuid",b"uuid"]) -> None: ...
global___Status = Status



import logging
import typing as t

import pyarrow as pa

from sarus_data_spec.manager.ops.base import (
    DatasetImplementation,
    DatasetStaticChecker,
    DataspecStaticChecker,
    ScalarImplementation,
)
from sarus_data_spec.manager.ops.processor.external.external_op import (  # noqa: E501
    ExternalDatasetOp,
    ExternalDatasetStaticChecker,
    ExternalOpImplementation,
    ExternalScalarOp,
    ExternalScalarStaticChecker,
    external_implementation,
)

try:
    from sarus_data_spec.manager.ops.processor.standard.differentiated_sample import (  # noqa: E501
        DifferentiatedSample,
        DifferentiatedSampleStaticChecker,
    )
except ModuleNotFoundError:
    logger = logging.getLogger(__name__)
    logger.info("Transforms: DifferentiatedSampling not available.")

from sarus_data_spec.manager.ops.processor.standard.extract import (
    Extract,
    ExtractStaticChecker,
)
from sarus_data_spec.manager.ops.processor.standard.filter import (
    Filter,
    FilterStaticChecker,
)
from sarus_data_spec.manager.ops.processor.standard.get_item import (
    GetItem,
    GetItemStaticChecker,
)
from sarus_data_spec.manager.ops.processor.standard.project import (
    Project,
    ProjectStaticChecker,
)

try:
    from sarus_data_spec.manager.ops.processor.standard.sample import (
        Sample,
        SampleStaticChecker,
    )

except ModuleNotFoundError:
    logger = logging.getLogger(__name__)
    logger.info("Transforms: Sample not available.")

try:
    from sarus_data_spec.manager.ops.processor.standard.select_sql import (
        SelectSQL,
        SelectSQLStaticChecker,
    )

except ModuleNotFoundError:
    logger = logging.getLogger(__name__)
    logger.info("Transforms: SelectSQL not available.")

try:
    from sarus_data_spec.manager.ops.processor.standard.dp_select_sql import (
        DPSelectSQL,
        DPSelectSQLStaticChecker,
    )

except ModuleNotFoundError:
    logger = logging.getLogger(__name__)
    logger.info("Transforms: SelectSQL not available.")


from sarus_data_spec.manager.ops.processor.standard.shuffle import (
    Shuffle,
    ShuffleStaticChecker,
)

try:
    from sarus_data_spec.manager.ops.processor.standard.protection_utils.protected_paths import (  # noqa: E501
        ProtectedPaths,
        ProtectedPathsStaticChecker,
        PublicPaths,
        PublicPathStaticChecker,
    )
    from sarus_data_spec.manager.ops.processor.standard.protection_utils.protection import (  # noqa: E501
        ProtectedDataset,
        ProtectedDatasetStaticChecker,
    )
except ModuleNotFoundError:
    logger = logging.getLogger(__name__)
    logger.info("Transforms: Protection not available.")

try:
    from sarus_data_spec.manager.ops.processor.standard.user_settings.automatic import (  # noqa: E501
        AutomaticUserSettings,
        AutomaticUserSettingsStaticChecker,
    )
    from sarus_data_spec.manager.ops.processor.standard.user_settings.user_settings import (  # noqa: E501
        UserSettingsDataset,
        UserSettingsStaticChecker,
    )
except ModuleNotFoundError:
    logger = logging.getLogger(__name__)
    logger.info("Transforms: UserSettings not available.")
try:
    from sarus_data_spec.manager.ops.processor.standard.assign_budget import (  # noqa: E501
        AssignBudget,
        AssignBudgetStaticChecker,
    )
    from sarus_data_spec.manager.ops.processor.standard.budgets_ops import (  # noqa: E501
        AttributesBudget,
        AttributesBudgetStaticChecker,
        AutomaticBudget,
        AutomaticBudgetStaticChecker,
        SDBudget,
        SDBudgetStaticChecker,
    )

except ModuleNotFoundError:
    logger = logging.getLogger(__name__)
    logger.info("Transforms: Transforms with budgets not available.")
try:
    from sarus_data_spec.manager.ops.processor.standard.synthetic import (  # noqa: E501
        SamplingRatios,
        SamplingRatiosStaticChecker,
        Synthetic,
        SyntheticStaticChecker,
    )
except ModuleNotFoundError:
    logger = logging.getLogger(__name__)
    logger.info("Transforms: Synthetic not available.")

try:
    from sarus_data_spec.manager.ops.processor.standard.derive_seed import (  # noqa: E501
        DeriveSeed,
        DeriveSeedStaticChecker,
    )
except ModuleNotFoundError:
    logger = logging.getLogger(__name__)
    logger.info("Transforms: Seed transforms not available.")
try:
    from sarus_data_spec.manager.ops.processor.standard.group_by_pe import (  # noqa: E501
        GroupByPE,
        GroupByPEStaticChecker,
    )
except ModuleNotFoundError:
    logger = logging.getLogger(__name__)
    logger.info("Transforms: GroupPE not available.")
try:
    from sarus_data_spec.manager.ops.processor.standard.transcode import (  # noqa: E501
        Transcode,
        TranscodeStaticChecker,
    )
except ModuleNotFoundError:
    logger = logging.getLogger(__name__)
    logger.info("Transforms: GroupPE not available.")
try:
    from sarus_data_spec.manager.ops.processor.standard.user_settings.relationship_spec import (  # noqa: E501
        RelationshipSpecOp,
        RelationshipSpecOpStaticChecker,
    )
except ModuleNotFoundError:
    logger = logging.getLogger(__name__)
    logger.info("Transforms: RelationshipSpec not available.")
try:
    from sarus_data_spec.manager.ops.processor.standard.user_settings.validated import (  # noqa: E501
        ValidatedUserTypeOp,
        ValidatedUserTypeOpStaticChecker,
    )
except ModuleNotFoundError:
    logger = logging.getLogger(__name__)
    logger.info("Transforms: ValidatedUserType not available.")
import sarus_data_spec.typing as st


def get_implementation(transform: st.Transform) -> ExternalOpImplementation:
    if not transform.is_external():
        raise NotImplementedError(
            "Cannot get implementation of internal transform."
        )
    return external_implementation(transform)


def get_dataset_op(
    transform: st.Transform,
) -> t.Tuple[t.Type[DatasetImplementation], t.Type[DatasetStaticChecker]]:
    if transform.is_external():
        return ExternalDatasetOp, ExternalDatasetStaticChecker
    elif transform.spec() == 'sample':
        return Sample, SampleStaticChecker
    elif transform.spec() == 'differentiated_sample':
        return DifferentiatedSample, DifferentiatedSampleStaticChecker
    elif transform.spec() == 'protect_dataset':
        return ProtectedDataset, ProtectedDatasetStaticChecker
    elif transform.spec() == 'user_settings':
        return UserSettingsDataset, UserSettingsStaticChecker
    elif transform.spec() == 'filter':
        return Filter, FilterStaticChecker
    elif transform.spec() == 'project':
        return Project, ProjectStaticChecker
    elif transform.spec() == 'shuffle':
        return Shuffle, ShuffleStaticChecker
    elif transform.spec() == 'synthetic':
        return Synthetic, SyntheticStaticChecker
    elif transform.spec() == 'get_item':
        return GetItem, GetItemStaticChecker
    elif transform.spec() == 'assign_budget':
        return AssignBudget, AssignBudgetStaticChecker
    elif transform.spec() == 'group_by_pe':
        return GroupByPE, GroupByPEStaticChecker
    elif transform.name() == 'Transcode':
        return Transcode, TranscodeStaticChecker
    elif transform.spec() == 'select_sql':
        return SelectSQL, SelectSQLStaticChecker
    elif transform.spec() == 'dp_select_sql':
        return DPSelectSQL, DPSelectSQLStaticChecker
    elif transform.spec() == 'extract':
        return Extract, ExtractStaticChecker
    else:
        raise NotImplementedError(transform.spec())


def get_scalar_op(
    transform: st.Transform,
) -> t.Tuple[t.Type[ScalarImplementation], t.Type[DataspecStaticChecker]]:
    if transform.is_external():
        return ExternalScalarOp, ExternalScalarStaticChecker
    elif transform.name() == 'automatic_protected_paths':
        # here we assume this transform is called
        # on a single dataset
        return ProtectedPaths, ProtectedPathsStaticChecker
    elif transform.name() == 'automatic_public_paths':
        # here we assume this transform is called
        # on a single dataset
        return PublicPaths, PublicPathStaticChecker
    elif transform.name() == 'automatic_user_settings':
        return AutomaticUserSettings, AutomaticUserSettingsStaticChecker
    elif transform.name() == 'automatic_budget':
        return AutomaticBudget, AutomaticBudgetStaticChecker
    elif transform.name() == 'attributes_budget':
        return AttributesBudget, AttributesBudgetStaticChecker
    elif transform.name() == 'sd_budget':
        return SDBudget, SDBudgetStaticChecker
    elif transform.name() == 'sampling_ratios':
        return SamplingRatios, SamplingRatiosStaticChecker
    elif transform.name() == 'derive_seed':
        return DeriveSeed, DeriveSeedStaticChecker
    elif transform.name() == 'relationship_spec':
        return RelationshipSpecOp, RelationshipSpecOpStaticChecker
    elif transform.name() == 'validated_user_type':
        return ValidatedUserTypeOp, ValidatedUserTypeOpStaticChecker
    else:
        raise NotImplementedError(f"scalar_transformed for {transform}")


class TransformedDataset(DatasetImplementation):
    def __init__(self, dataset: st.Dataset):
        super().__init__(dataset)
        transform = self.dataset.transform()
        ImplementationClass, StaticCheckerClass = get_dataset_op(transform)
        self.implementation = ImplementationClass(dataset)
        self.static_checker = StaticCheckerClass(dataset)

    async def private_queries(self) -> t.List[st.PrivateQuery]:
        return await self.static_checker.private_queries()

    async def to_arrow(
        self, batch_size: int
    ) -> t.AsyncIterator[pa.RecordBatch]:
        return await self.implementation.to_arrow(batch_size)

    async def schema(self) -> st.Schema:
        return await self.static_checker.schema()

    async def size(self) -> st.Size:
        return await self.implementation.size()

    async def multiplicity(self) -> st.Multiplicity:
        return await self.implementation.multiplicity()

    async def bounds(self) -> st.Bounds:
        return await self.implementation.bounds()

    async def marginals(self) -> st.Marginals:
        return await self.implementation.marginals()

    async def sql(
        self,
        query: t.Union[str, t.Dict[str, t.Any]],
        dialect: t.Optional[st.SQLDialect] = None,
        batch_size: int = 10000,
        result_type: t.Optional[st.Type] = None,
    ) -> t.AsyncIterator[pa.RecordBatch]:
        return await self.implementation.sql(
            query, dialect, batch_size, result_type
        )


class TransformedScalar(ScalarImplementation):
    def __init__(self, scalar: st.Scalar):
        super().__init__(scalar)
        transform = self.scalar.transform()
        ImplementationClass, StaticCheckerClass = get_scalar_op(transform)
        self.implementation = ImplementationClass(scalar)
        self.static_checker = StaticCheckerClass(scalar)

    async def value(self) -> t.Any:
        return await self.implementation.value()

    async def private_queries(self) -> t.List[st.PrivateQuery]:
        return await self.static_checker.private_queries()



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: sarus_data_spec/protobuf/links.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from sarus_data_spec.protobuf import path_pb2 as sarus__data__spec_dot_protobuf_dot_path__pb2
from sarus_data_spec.protobuf import statistics_pb2 as sarus__data__spec_dot_protobuf_dot_statistics__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n$sarus_data_spec/protobuf/links.proto\x12\x0fsarus_data_spec\x1a#sarus_data_spec/protobuf/path.proto\x1a)sarus_data_spec/protobuf/statistics.proto\"\xdb\x03\n\x05Links\x12\x0c\n\x04uuid\x18\x01 \x01(\t\x12\x0f\n\x07\x64\x61taset\x18\x02 \x01(\t\x12\x39\n\x10links_statistics\x18\x03 \x03(\x0b\x32\x1f.sarus_data_spec.Links.LinkStat\x12:\n\nproperties\x18\x04 \x03(\x0b\x32&.sarus_data_spec.Links.PropertiesEntry\x1a\x31\n\x0fPropertiesEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01\x1a\x88\x02\n\x08LinkStat\x12\'\n\x08pointing\x18\x01 \x01(\x0b\x32\x15.sarus_data_spec.Path\x12&\n\x07pointed\x18\x02 \x01(\x0b\x32\x15.sarus_data_spec.Path\x12\x33\n\x0c\x64istribution\x18\x03 \x01(\x0b\x32\x1d.sarus_data_spec.Distribution\x12\x43\n\nproperties\x18\x04 \x03(\x0b\x32/.sarus_data_spec.Links.LinkStat.PropertiesEntry\x1a\x31\n\x0fPropertiesEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01\x62\x06proto3')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'sarus_data_spec.protobuf.links_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  _LINKS_PROPERTIESENTRY._options = None
  _LINKS_PROPERTIESENTRY._serialized_options = b'8\001'
  _LINKS_LINKSTAT_PROPERTIESENTRY._options = None
  _LINKS_LINKSTAT_PROPERTIESENTRY._serialized_options = b'8\001'
  _LINKS._serialized_start=138
  _LINKS._serialized_end=613
  _LINKS_PROPERTIESENTRY._serialized_start=297
  _LINKS_PROPERTIESENTRY._serialized_end=346
  _LINKS_LINKSTAT._serialized_start=349
  _LINKS_LINKSTAT._serialized_end=613
  _LINKS_LINKSTAT_PROPERTIESENTRY._serialized_start=297
  _LINKS_LINKSTAT_PROPERTIESENTRY._serialized_end=346
# @@protoc_insertion_point(module_scope)



import typing as t

from sarus_data_spec.context.typing import Context

_global_context_stack: t.List[Context] = []


def push_global_context(context: Context) -> None:
    _global_context_stack.append(context)


def pop_global_context() -> Context:
    return _global_context_stack.pop()


def global_context() -> Context:
    return _global_context_stack[-1]



from __future__ import annotations

import hashlib
import typing as t

import pandas as pd
import pyarrow as pa

from sarus_data_spec.arrow.admin_utils import (
    compute_admin_data,
    create_admin_columns,
    merge_data_and_admin,
)
from sarus_data_spec.arrow.conversion import to_pyarrow_table
from sarus_data_spec.arrow.schema import type_from_arrow_schema
from sarus_data_spec.dataspec_validator.signature import (
    SarusBoundSignature,
    SarusSignature,
    SarusSignatureValue,
)
from sarus_data_spec.dataspec_validator.typing import PEPKind
from sarus_data_spec.manager.async_utils import async_iter
from sarus_data_spec.manager.ops.base import (
    DatasetImplementation,
    DatasetStaticChecker,
    DataspecStaticChecker,
    ScalarImplementation,
)
from sarus_data_spec.schema import schema as schema_builder
from sarus_data_spec.transform import external, transform_id
import sarus_data_spec.protobuf as sp
import sarus_data_spec.type as sdt
import sarus_data_spec.typing as st

from .typing import NO_TRANSFORM_ID
from .utils import static_arguments


class ExternalScalarStaticChecker(DataspecStaticChecker):
    async def private_queries(self) -> t.List[st.PrivateQuery]:
        """Return the PrivateQueries summarizing DP characteristics."""
        implementation = external_implementation(self.dataspec.transform())
        bound_signature = implementation.signature().bind_dataspec(
            self.dataspec
        )
        return await implementation.private_queries(bound_signature)

    def is_dp(self) -> bool:
        """Checks if the transform is DP and compatible with the arguments."""
        implementation = external_implementation(self.dataspec.transform())
        bound_signature = implementation.signature().bind_dataspec(
            self.dataspec
        )
        return implementation.is_dp(bound_signature)

    def is_dp_applicable(self, public_context: t.Collection[str]) -> bool:
        """Statically check if a DP transform is applicable in this position.

        This verification is common to all dataspecs and is true if:
            - the dataspec is transformed and its transform has an equivalent
            DP transform
            - the DP transform's required PEP arguments are PEP and aligned
            (i.e. same PEP token)
            - other dataspecs arguments are public
        """
        transform = self.dataspec.transform()
        implementation = external_implementation(transform)
        dp_implementation = implementation.dp_equivalent()
        bound_signature = implementation.signature().bind_dataspec(
            self.dataspec
        )
        if dp_implementation is None or not dp_implementation.is_dp(
            bound_signature
        ):
            return False

        return bound_signature.pep_token() is not None

    def dp_transform(self) -> t.Optional[st.Transform]:
        """Return the dataspec's DP equivalent transform if existing."""
        transform = self.dataspec.transform()
        op_implementation = external_implementation(transform)
        py_args, py_kwargs, ds_args_pos, ds_types = static_arguments(transform)

        dp_implementation = op_implementation.dp_equivalent()
        if dp_implementation is None:
            return None

        dp_transform_id = dp_implementation.transform_id()
        assert dp_transform_id is not None

        # Types won't be used since budget & seed are scalars
        ds_types["budget"] = ""
        ds_types["seed"] = ""

        return external(
            dp_transform_id,
            py_args=py_args,
            py_kwargs=py_kwargs,
            ds_args_pos=ds_args_pos,
            ds_types=ds_types,
        )


class ExternalDatasetStaticChecker(
    ExternalScalarStaticChecker, DatasetStaticChecker
):
    def __init__(self, dataset: st.Dataset):
        super().__init__(dataset)
        self.dataset = dataset

    def pep_token(self, public_context: t.Collection[str]) -> t.Optional[str]:
        """Return the dataspec's PEP token."""
        transform = self.dataspec.transform()
        implementation = external_implementation(transform)
        bound_signature = implementation.signature().bind_dataspec(
            self.dataspec
        )

        input_token = bound_signature.pep_token()
        if input_token is None:
            return None

        pep_kind = implementation.pep_kind(bound_signature)
        if pep_kind == PEPKind.NOT_PEP:
            return None
        elif pep_kind == PEPKind.TOKEN_PRESERVING:
            return input_token
        else:  # PEP or ROW
            h = hashlib.md5()
            h.update(input_token.encode("ascii"))
            h.update(transform.protobuf().SerializeToString())
            new_token = h.hexdigest()
            return new_token

    async def schema(self) -> st.Schema:
        """Computes the schema of the dataspec.

        The schema is computed by computing the synthetic data value and
        converting the Pyarrow schema to a Sarus schema.q
        """
        syn_variant = self.dataset.variant(kind=st.ConstraintKind.SYNTHETIC)
        assert syn_variant is not None
        assert syn_variant.prototype() == sp.Dataset

        syn_dataset = t.cast(st.Dataset, syn_variant)
        arrow_iterator = await syn_dataset.async_to_arrow(batch_size=1)
        first_batch = await arrow_iterator.__anext__()
        schema = first_batch.schema

        schema_type = type_from_arrow_schema(schema)
        if self.dataset.is_pep():
            # If the dataset is PEP then the schema of the real data should
            # have protection but the synthetic data might not have it. We
            # need to add it manually.
            schema_type = sdt.protected_type(schema_type)

        return schema_builder(self.dataset, schema_type=schema_type)


class ExternalDatasetOp(DatasetImplementation):
    async def to_arrow(
        self, batch_size: int
    ) -> t.AsyncIterator[pa.RecordBatch]:
        transform = self.dataset.transform()
        implementation = external_implementation(transform)
        bound_signature = implementation.signature().bind_dataspec(
            self.dataset
        )
        bound_signature.static_validation()

        if self.dataset.is_pep():
            result = await implementation.compute(bound_signature)
            if (
                isinstance(result, pd.Series)
                and implementation.pep_kind(bound_signature) == PEPKind.ROW
            ):
                # Reformat the series as a dataframe with a single row
                result = result.to_frame().transpose()

            ds_result = t.cast(st.DatasetCastable, result)
            admin_data = await bound_signature.admin_data()
            output_admin_data = compute_admin_data(admin_data, ds_result)
            data_table = to_pyarrow_table(ds_result)
            table = merge_data_and_admin(data_table, output_admin_data)

        else:
            result = await implementation.compute(bound_signature)
            data_table = to_pyarrow_table(result)
            table = create_admin_columns(data_table)

        return async_iter(table.to_batches(max_chunksize=batch_size))


class ExternalScalarOp(ScalarImplementation):
    async def value(self) -> t.Any:
        transform = self.scalar.transform()
        ds_args, ds_kwargs = self.scalar.parents()
        return await async_compute_external_value(
            transform, *ds_args, **ds_kwargs
        )


async def async_compute_external_value(
    transform: st.Transform,
    *ds_args: t.Union[st.DataSpec, st.Transform],
    **ds_kwargs: t.Union[st.DataSpec, st.Transform],
) -> t.Any:
    """Compute the value of an external transform applied on Dataspecs.

    This function computes the output value without manipulating the
    corresponding Dataspec. This is useful when we need to have access
    to the value of a Dataspec before its creation:
      - for computing a Mock value and inferring if the result is
        a Scalar or a Dataset.
    """
    implementation = external_implementation(transform)
    bound_signature = implementation.signature().bind(
        transform, *ds_args, **ds_kwargs
    )
    bound_signature.static_validation()
    data = await implementation.compute(bound_signature)
    return data


class ExternalOpImplementation:
    """External PEP op implementation class.

    This class wraps together several elements of an external op
    implementation:
        - `call` is the function that computes the output value from the
          input(s) value(s).
    """

    _transform_id: str = NO_TRANSFORM_ID
    _dp_equivalent_id: t.Optional[str] = None
    _non_dp_equivalent_id: t.Optional[str] = None
    _signature: t.Optional[SarusSignature] = None

    def transform_id(self) -> str:
        return self._transform_id

    def dp_equivalent_id(self) -> t.Optional[str]:
        return self._dp_equivalent_id

    def dp_equivalent(self) -> t.Optional[ExternalOpImplementation]:
        if not self._dp_equivalent_id:
            return None

        return external_implementation_from_id(self._dp_equivalent_id)

    def signature(self) -> SarusSignature:
        if self._signature is not None:
            return self._signature

        if self._non_dp_equivalent_id is None:
            raise ValueError(
                f"External implementation {self.transform_id()} has no "
                "signature defined and no non-DP equivalent."
            )

        non_dp_signature = external_implementation_from_id(
            self._non_dp_equivalent_id
        ).signature()
        return non_dp_signature.make_dp()

    async def compute(self, bound_signature: SarusBoundSignature) -> t.Any:
        if self.is_dp(bound_signature):
            return await self.call_dp(bound_signature)
        else:
            signature_value = await bound_signature.collect_signature_value()
            return self.call(signature_value)

    def call(self, signature_value: SarusSignatureValue) -> t.Any:
        raise NotImplementedError

    async def call_dp(self, bound_signature: SarusBoundSignature) -> t.Any:
        """DP ops `call` need to be async to compute schema, tasks, etc"""
        raise NotImplementedError

    def pep_kind(self, bound_signature: SarusBoundSignature) -> PEPKind:
        """Return the PEP properties of the transform.

        It takes the transform arguments as input because it can depend on some
        transform parameters. For instance, it is not PEP if we are aggregating
        the rows (axis=0) and it is PEP if we are aggregating the columns
        (axis=1).
        """
        # Default implementation
        return PEPKind.NOT_PEP

    def is_dp(self, bound_signature: SarusBoundSignature) -> bool:
        """Return True if the DP transform is compatible with the arguments.

        It takes the transform arguments as input because it can depend on some
        transform parameters. For instance, if we are aggregating the rows
        (axis=0), then there might be an equivalent DP transform but if we are
        aggregating the columns there might not (axis=1).
        """
        # Default implementation
        return False

    async def private_queries(
        self, signature: SarusBoundSignature
    ) -> t.List[st.PrivateQuery]:
        """Takes as input the args of the transform (static and dynamic)."""
        if not signature.is_dp():
            return []
        queries, _ = await self.private_queries_and_task(signature)
        return queries

    async def private_queries_and_task(
        self, signature: SarusBoundSignature
    ) -> t.Tuple[t.List[st.PrivateQuery], st.Task]:
        raise NotImplementedError

    def callable(
        self, composed_transform: st.Transform
    ) -> t.Callable[..., t.Awaitable[t.Any]]:
        """Build the transform's async callable.

        The function takes an undefined number of named arguments.

        It first collects the current transform's signature concrete values
        using the passed variables' values. The concrete values are stored in a
        SarusBoundSignature object so we can compute the current transform's
        output by simply using the implementation's `call` method.
        """
        lambda_signature = self.signature().bind_composed(composed_transform)
        previous_callable = lambda_signature.callable()

        def composed_callable(*vars: t.Any, **kwvars: t.Any) -> t.Any:
            signature_value = previous_callable(*vars, **kwvars)
            return self.call(signature_value)

        return composed_callable


def external_implementation(
    transform: st.Transform,
) -> ExternalOpImplementation:
    """Return the implementation of an external op from a DataSpec.

    The mapping is done by the config file.
    """
    assert transform and transform.is_external()
    id = transform_id(transform)
    return external_implementation_from_id(id)


def external_implementation_from_id(id: str) -> ExternalOpImplementation:
    # Imported here to avoid circular imports
    from . import ROUTING

    library, op_name = id.split(".")
    op_implementation = t.cast(
        ExternalOpImplementation, ROUTING[library][op_name]
    )
    return op_implementation
